#!/usr/bin/env python

import os
import math
from contextlib import suppress

from mapel.elections.features.dependent_rounding import approx_rand_tree

try:
    import cplex
except ImportError:
    cplex = None

import numpy as np


# NEW ILP
def solve_rand_approx_pav(election, committee_size, W, C, ctr=0, fixed=None):
    if ctr == 1:
        return 0

    if fixed is None:
        fixed = []

    cp = cplex.Cplex()
    cp.parameters.threads.set(1)

    m = election.num_candidates
    n = election.num_voters
    k = committee_size

    # OBJECTIVE FUNCTION
    cp.objective.set_sense(cp.objective.sense.minimize)
    names = []
    obj = []
    for j in range(n):
        for l in range(k):
            for i in range(m):
                names.append(f'x{l}_{i}_{j}')
                obj.append(W[l] * C[j][i])
    types = [cp.variables.type.continuous] * len(names)
    cp.variables.add(obj=obj, names=names, types=types,
                     lb=[0.] * len(names), ub=[1.] * len(names))

    # ADD MISSING VARIABLES
    names = []
    for i in range(m):
        names.append(f'y{i}')
    cp.variables.add(names=list(names),
                     types=[cp.variables.type.continuous] * len(names),
                     lb=[0.] * len(names), ub=[1.] * len(names))

    # FIRST SINGLE CONSTRAINT
    ind = []
    val = []
    for i in range(m):
        ind.append(f'y{i}')
        val.append(1.)
    cp.linear_constraints.add(lin_expr=[cplex.SparsePair(ind=ind, val=val)],
                              senses=['E'],
                              rhs=[committee_size],
                              names=['C1'])

    # SECOND GROUP OF CONSTRAINTS
    lin_expr = []
    rhs = []
    for i in range(m):
        for j in range(n):
            val = []
            ind = []
            for l in range(k):
                ind.append(f'x{l}_{i}_{j}')
                val.append(1.)
            ind.append(f'y{i}')
            val.append(-1.)
            lin_expr.append(cplex.SparsePair(ind=ind, val=val))
            rhs.append(0.)
    cp.linear_constraints.add(lin_expr=lin_expr,
                              senses=['L'] * len(rhs),
                              rhs=rhs,
                              names=['C2_' + str(i) for i in range(len(rhs))])

    # THIRD GROUP OF CONSTRAINTS
    lin_expr = []
    rhs = []
    for j in range(n):
        for l in range(k):
            val = []
            ind = []
            for i in range(m):
                ind.append(f'x{l}_{i}_{j}')
                val.append(1.)
            lin_expr.append(cplex.SparsePair(ind=ind, val=val))
            rhs.append(1.)
    cp.linear_constraints.add(lin_expr=lin_expr,
                              senses=['G'] * len(rhs),
                              rhs=rhs,
                              names=['C3_' + str(i) for i in range(len(rhs))])

    # FIXED
    lin_expr = []
    rhs = []
    for i in fixed:
        ind = [f'y{i}']
        val = [1.]
    lin_expr.append(cplex.SparsePair(ind=ind, val=val))
    rhs.append(1.)
    cp.linear_constraints.add(lin_expr=lin_expr,
                              senses=['E'] * len(rhs),
                              rhs=rhs,
                              names=['C4_' + str(i) for i in range(len(rhs))])

    cp.set_results_stream(None)
    try:
        cp.solve()
    except:  # cplex.CplexSolverError:
        print("Exception raised while solving")
        return

    ### TO DO LATER ###

    # score, fixed = print_results(election, committee_size, cp)

    values = [0.] * election.num_candidates
    for i in range(election.num_candidates):
        values[i] = cp.solution.get_values('y' + str(i))

    final_values = approx_rand_tree(values)

    winner_id = 0
    winners = [-1] * committee_size
    for i in range(election.num_candidates):
        if final_values[f'x{i}'] == 1.:
            winners[winner_id] = i
            winner_id += 1
    winners = sorted(winners)

    def get_pav_score(election, winners) -> float:

        num_voters = election.num_voters
        num_candidates = election.num_candidates
        votes = election.votes

        score = 0

        vector = [0.] * num_candidates
        for i in range(len(winners)):
            vector[i] = 1.

        for i in range(num_voters):
            ctr = 1.
            for j in range(num_candidates):
                if votes[i][j] in winners:
                    score += (1. / ctr) * vector[j]
                    ctr += 1

        return score

    try:
        score = get_pav_score(election, winners)
    except:
        score = -1

    return score


# FOR SUBELECTIONS
def solve_lp_voter_subelection(election_1, election_2, metric_name='0'):
    """ LP solver for voter subelection problem """

    cp = cplex.Cplex()
    cp.parameters.threads.set(1)

    # OBJECTIVE FUNCTION
    cp.objective.set_sense(cp.objective.sense.maximize)
    names = []
    for v1 in range(election_1.num_voters):
        for v2 in range(election_2.num_voters):
            names.append('N' + str(v1) + '_' + str(v2))
    obj = [1.] * len(names)
    types = [cp.variables.type.binary] * len(names)
    cp.variables.add(obj=obj, names=names, types=types)

    # FIRST CONSTRAINT FOR VOTERS
    lin_expr = []
    for v1 in range(election_1.num_voters):
        val = [1.0] * election_2.num_voters
        ind = []
        for v2 in range(election_2.num_voters):
            ind.append('N' + str(v1) + '_' + str(v2))
        lin_expr.append(cplex.SparsePair(ind=ind, val=val))
    cp.linear_constraints.add(lin_expr=lin_expr,
                              senses=['L'] * election_1.num_voters,
                              rhs=[1.0] * election_1.num_voters,
                              names=['C1_' + str(i) for i in range(election_1.num_voters)])

    # SECOND CONSTRAINT FOR VOTERS
    lin_expr = []
    for v2 in range(election_2.num_voters):
        ind = []
        for v1 in range(election_1.num_voters):
            ind.append('N' + str(v1) + '_' + str(v2))
        lin_expr.append(cplex.SparsePair(ind=ind, val=[1.0] * election_1.num_voters))
    cp.linear_constraints.add(lin_expr=lin_expr,
                              senses=['L'] * election_2.num_voters,
                              rhs=[1.0] * election_2.num_voters,
                              names=['C2_' + str(i) for i in range(election_2.num_voters)])

    # ADD VARIABLES FOR CANDIDATES
    names = []
    for c1 in range(election_1.num_candidates):
        for c2 in range(election_2.num_candidates):
            names.append('M' + str(c1) + '_' + str(c2))
    cp.variables.add(names=list(names),
                     types=[cp.variables.type.binary] * election_1.num_candidates * election_2.num_candidates)

    # FIRST CONSTRAINT FOR CANDIDATES
    lin_expr = []
    for c1 in range(election_1.num_candidates):
        ind = []
        for c2 in range(election_2.num_candidates):
            ind.append('M' + str(c1) + '_' + str(c2))
        lin_expr.append(cplex.SparsePair(ind=ind, val=[1.0] * election_2.num_candidates))
    cp.linear_constraints.add(lin_expr=lin_expr,
                              senses=['E'] * election_1.num_candidates,
                              rhs=[1.0] * election_1.num_candidates,
                              names=['C3_' + str(i) for i in range(election_1.num_candidates)])

    # SECOND CONSTRAINT FOR CANDIDATES
    lin_expr = []
    for c2 in range(election_2.num_candidates):
        ind = []
        for c1 in range(election_1.num_candidates):
            ind.append('M' + str(c1) + '_' + str(c2))
        lin_expr.append(cplex.SparsePair(ind=ind, val=[1.0] * election_1.num_candidates))
    cp.linear_constraints.add(lin_expr=lin_expr,
                              senses=['E'] * election_2.num_candidates,
                              rhs=[1.0] * election_2.num_candidates,
                              names=['C4_' + str(i) for i in range(election_2.num_candidates)])


    # MAIN CONSTRAINT FOR VOTES
    potes_1 = election_1.get_potes()
    potes_2 = election_2.get_potes()
    lin_expr = []
    for v1 in range(election_1.num_voters):
        for v2 in range(election_2.num_voters):
            ind = []
            val = []
            for c1 in range(election_1.num_candidates):
                for c2 in range(election_2.num_candidates):
                    ind.append('M' + str(c1) + '_' + str(c2))
                    if abs(potes_1[v1][c1] - potes_2[v2][c2]) <= int(metric_name):
                        val.append(1.)
                    else:
                        val.append(0.)
            ind.append('N' + str(v1) + '_' + str(v2))
            val.append(-election_1.num_candidates)
            lin_expr.append(cplex.SparsePair(ind=ind, val=val))
    cp.linear_constraints.add(lin_expr=lin_expr,
                              senses=['G'] * election_1.num_voters * election_2.num_voters,
                              rhs=[0.0] * election_1.num_voters * election_2.num_voters,
                              names=['C5_' + str(i) for i in
                                     range(election_1.num_voters * election_2.num_voters)])

    # cp.write('new.lp')

    # SOLVE THE ILP
    cp.set_results_stream(None)
    try:
        cp.solve()
    except:  # cplex.CplexSolverError:
        print("Exception raised while solving")
        return

    return cp.solution.get_objective_value()


def solve_lp_candidate_subelections(lp_file_name, election_1, election_2):
    """ LP solver for candidate subelection problem """

    # PRECOMPUTING
    # """

    P = np.zeros([election_1.num_voters, election_2.num_voters, election_1.num_candidates,
                  election_2.num_candidates,
                  election_1.num_candidates, election_2.num_candidates])

    for v in range(election_1.num_voters):
        for u in range(election_2.num_voters):
            for c1 in range(election_1.num_candidates):
                for d1 in range(election_2.num_candidates):
                    for c2 in range(election_1.num_candidates):
                        for d2 in range(election_2.num_candidates):
                            if (election_1.potes[v][c1] > election_1.potes[v][c2] and
                                election_2.potes[u][d1] >
                                election_2.potes[u][d2]) or \
                                    (election_1.potes[v][c1] < election_1.potes[v][c2] and
                                     election_2.potes[u][d1] <
                                     election_2.potes[u][d2]):
                                P[v][u][c1][d1][c2][d2] = 1


    # CREATE LP FILE
    lp_file = open(lp_file_name, 'w')
    lp_file.write("Maximize\nobj: ")

    first = True
    for c in range(election_1.num_candidates):
        for d in range(election_2.num_candidates):
            if not first:
                lp_file.write(" + ")
            first = False
            lp_file.write(" M_" + str(c) + "_" + str(d))

    lp_file.write("\n")

    """
    first = True
    for v in range(election_1.num_voters):
        for u in range(election_2.num_voters):
            for c1 in range(election_1.num_candidates):
                for d1 in range(election_2.num_candidates):
                    for c2 in range(election_1.num_candidates):
                        if c1 == c2:
                            continue
                        for d2 in range(election_2.num_candidates):
                            if d1 == d2:
                                continue

                            if P[v][u][c1][d1][c2][d2] == 1:
                                if not first:
                                    lp_file.write(" + ")
                                first = False
                                lp_file.write(" P_" + str(v) + "_" + str(u) + "_" +
                                              str(c1) + "_" + str(d1) + "_" + str(c2) + "_" + str(d2))
    lp_file.write("\n")
    """

    lp_file.write("Subject To\n")
    ctr_c = 0

    # FIRST CONSTRAINT FOR VOTERS
    for v in range(election_1.num_voters):
        lp_file.write("c" + str(ctr_c) + ":")
        first = True
        for u in range(election_2.num_voters):
            if not first:
                lp_file.write(" +")
            first = False
            lp_file.write(" N_" + str(v) + "_" + str(u))
        lp_file.write(" = 1" + "\n")
        ctr_c += 1

    # SECOND CONSTRAINT FOR VOTERS
    for u in range(election_2.num_voters):
        lp_file.write("c" + str(ctr_c) + ":")
        first = True
        for v in range(election_1.num_voters):
            if not first:
                lp_file.write(" +")
            first = False
            lp_file.write(" N_" + str(v) + "_" + str(u))
        lp_file.write(" = 1" + "\n")
        ctr_c += 1

    # FIRST CONSTRAINT FOR CANDIDATES
    for c in range(election_1.num_candidates):
        lp_file.write("c" + str(ctr_c) + ":")
        first = True
        for d in range(election_2.num_candidates):
            if not first:
                lp_file.write(" +")
            first = False
            lp_file.write(" M_" + str(c) + "_" + str(d))
        lp_file.write(" <= 1" + "\n")
        ctr_c += 1

    # SECOND CONSTRAINT FOR CANDIDATES
    for d in range(election_2.num_candidates):
        lp_file.write("c" + str(ctr_c) + ":")
        first = True
        for c in range(election_1.num_candidates):
            if not first:
                lp_file.write(" +")
            first = False
            lp_file.write(" M_" + str(c) + "_" + str(d))
        lp_file.write(" <= 1" + "\n")
        ctr_c += 1

    # FIRST CONSTRAINT FOR P
    for v in range(election_1.num_voters):
        for u in range(election_2.num_voters):
            for c1 in range(election_1.num_candidates):
                for d1 in range(election_2.num_candidates):
                    for c2 in range(election_1.num_candidates):
                        if c1 == c2:
                            continue
                        for d2 in range(election_2.num_candidates):
                            if d1 == d2:
                                continue

                            # if P[v][u][c1][d1][c2][d2] == 1:
                            lp_file.write("c" + str(ctr_c) + ":")
                            lp_file.write(" P_" + str(v) + "_" + str(u) + "_" +
                                          str(c1) + "_" + str(d1) + "_" + str(c2) + "_" + str(d2))

                            lp_file.write(" - 0.34 N_" + str(v) + "_" + str(u))
                            lp_file.write(" - 0.34 M_" + str(c1) + "_" + str(d1))
                            lp_file.write(" - 0.34 M_" + str(c2) + "_" + str(d2))

                            lp_file.write(" <= 0" + "\n")
                            ctr_c += 1

    # SECOND CONSTRAINT FOR P
    for v in range(election_1.num_voters):
        for u in range(election_2.num_voters):
            for c1 in range(election_1.num_candidates):
                for d1 in range(election_2.num_candidates):
                    for c2 in range(election_1.num_candidates):
                        if c1 == c2:
                            continue
                        for d2 in range(election_2.num_candidates):
                            if d1 == d2:
                                continue

                            # if P[v][u][c1][d1][c2][d2] == 1:
                            lp_file.write("c" + str(ctr_c) + ":")
                            lp_file.write(" P_" + str(v) + "_" + str(u) + "_" +
                                          str(c1) + "_" + str(d1) + "_" + str(c2) + "_" + str(d2))
                            # lp_file.write(" + 1")
                            lp_file.write(" - 0.34 N_" + str(v) + "_" + str(u))
                            lp_file.write(" - 0.34 M_" + str(c1) + "_" + str(d1))
                            lp_file.write(" - 0.34 M_" + str(c2) + "_" + str(d2))

                            lp_file.write(" > -1" + "\n")
                            ctr_c += 1

    # THIRD CONSTRAINT FOR P
    for v in range(election_1.num_voters):
        for u in range(election_2.num_voters):
            for c1 in range(election_1.num_candidates):
                for d1 in range(election_2.num_candidates):
                    for c2 in range(election_1.num_candidates):
                        if c1 == c2:
                            continue
                        for d2 in range(election_2.num_candidates):
                            if d1 == d2:
                                continue

                            # if P[v][u][c1][d1][c2][d2] == 1:
                            lp_file.write("c" + str(ctr_c) + ":")
                            lp_file.write(" P_" + str(v) + "_" + str(u) + "_" +
                                          str(c1) + "_" + str(d1) + "_" + str(c2) + "_" + str(d2))

                            lp_file.write(" <= " + str(P[v][u][c1][d1][c2][d2]) + "\n")
                            ctr_c += 1

    """
    # NEW 1
    for c1 in range(election_1.num_candidates):
        for d1 in range(election_2.num_candidates):
            lp_file.write("c" + str(ctr_c) + ":")
            first = True
            for v in range(election_1.num_voters):
                for u in range(election_2.num_voters):
                    for c2 in range(election_1.num_candidates):
                        if c1 == c2:
                            continue
                        for d2 in range(election_2.num_candidates):
                            if d1 == d2:
                                continue

                            if P[v][u][c1][d1][c2][d2] == 1:
                                if not first:
                                    lp_file.write(" +")
                                first = False
                                lp_file.write(" P_" + str(v) + "_" + str(u) + "_" +
                                              str(c1) + "_" + str(d1) + "_" + str(c2) + "_" + str(d2))
            lp_file.write(' - ' + str((magic_param-1)*election_1.num_voters) + ' M_' + str(c1) + '_' + str(d1) + ' = 0' + "\n")
            ctr_c += 1

    # NEW 2
    for v in range(election_1.num_voters):
        for u in range(election_2.num_voters):
            lp_file.write("c" + str(ctr_c) + ":")
            first = True
            for c1 in range(election_1.num_candidates):
                for d1 in range(election_2.num_candidates):
                    for c2 in range(election_1.num_candidates):
                        if c1 == c2:
                            continue
                        for d2 in range(election_2.num_candidates):
                            if d1 == d2:
                                continue

                            if P[v][u][c1][d1][c2][d2] == 1:
                                if not first:
                                    lp_file.write(" +")
                                first = False
                                lp_file.write(" P_" + str(v) + "_" + str(u) + "_" +
                                              str(c1) + "_" + str(d1) + "_" + str(c2) + "_" + str(d2))
            lp_file.write(' - ' + str((magic_param-1)*2) + ' N_' + str(v) + '_' + str(u) + ' = 0' + "\n")
            ctr_c += 1
    """

    lp_file.write("Binary\n")

    for v in range(election_1.num_voters):
        for u in range(election_2.num_voters):
            for c1 in range(election_1.num_candidates):
                for d1 in range(election_2.num_candidates):
                    for c2 in range(election_1.num_candidates):
                        if c1 == c2:
                            continue
                        for d2 in range(election_2.num_candidates):
                            if d1 == d2:
                                continue

                            # if P[v][u][c1][d1][c2][d2] == 1:
                            lp_file.write("P_" + str(v) + "_" + str(u) + "_" +
                                          str(c1) + "_" + str(d1) + "_" + str(c2) + "_" + str(
                                d2) + "\n")

    for v in range(election_1.num_voters):
        for u in range(election_2.num_voters):
            lp_file.write("N_" + str(v) + "_" + str(u) + "\n")

    for c in range(election_1.num_candidates):
        for d in range(election_2.num_candidates):
            lp_file.write("M_" + str(c) + "_" + str(d) + "\n")

    lp_file.write("End\n")

    lp_file.close()

    ### SECOND PART
    cp_lp = cplex.Cplex(lp_file_name)
    cp_lp.parameters.threads.set(1)
    cp_lp.set_results_stream(None)

    try:
        cp_lp.solve()
    except:  # cplex.CplexSolverError:
        print("Exception raised during solve")
        return

    ##########################
    ##########################

    result = np.zeros([election_1.num_candidates, election_1.num_candidates])
    for i in range(election_1.num_candidates):
        for j in range(election_1.num_candidates):
            name = 'M_' + str(i) + '_' + str(j)
            result[i][j] = cp_lp.solution.get_values(name)

    # print('M', result)
    """
    result_2 = np.zeros([election_1.num_voters, election_1.num_voters])
    for i in range(election_1.num_voters):
        for j in range(election_1.num_voters):
            election_id = 'N_' + str(i) + '_' + str(j)
            result_2[i][j] = cp_lp.solution.get_values(election_id)

    print('N', result_2)
    total = 0
    for v in range(election_1.num_voters):
        for u in range(election_1.num_voters):
            for c1 in range(election_1.num_candidates):
                for d1 in range(election_1.num_candidates):
                    for c2 in range(election_1.num_candidates):
                        if c1 == c2:
                            continue
                        for d2 in range(election_1.num_candidates):
                            if d1 == d2:
                                continue

                            #if P[v][u][c1][d1][c2][d2] == 1:
                            election_id = "P_" + str(v) + "_" + str(u) + "_" + str(c1) + "_" + str(d1) + "_" + str(c2) + "_" + str(d2)
                            value = cp_lp.solution.get_values(election_id)
                            #print(value)
                            if value == 1:
                                print(election_id)
                            total += value
    print(total)
    """
    ##########################
    ##########################

    # objective_value = cp_lp.solution.get_objective_value()
    return sum(sum(result))


# FOR METRICS
def solve_lp_matching_vector_with_lp(cost_table, length):
    """ LP solver for vectors' matching """

    # print(cost_table)
    cp = cplex.Cplex()
    cp.parameters.threads.set(1)

    # OBJECTIVE FUNCTION
    cp.objective.set_sense(cp.objective.sense.minimize)
    objective = []
    names = []
    pos = 0
    for i in range(length):
        for j in range(length):
            names.append('x' + str(pos))
            objective.append(cost_table[i][j])
            pos += 1
    cp.variables.add(obj=objective, names=names, types=[cp.variables.type.binary] * length ** 2)

    # FIRST GROUP OF CONSTRAINTS
    lin_expr = []
    for i in range(length):
        ind = []
        for j in range(length):
            pos = i * length + j
            ind.append('x' + str(pos))
        lin_expr.append(cplex.SparsePair(ind=ind, val=[1.0] * length))
    cp.linear_constraints.add(lin_expr=lin_expr, senses=['E'] * length, rhs=[1.0] * length)

    # SECOND GROUP OF CONSTRAINTS
    lin_expr = []
    for j in range(length):
        ind = []
        for i in range(length):
            pos = i * length + j
            ind.append('x' + str(pos))
        lin_expr.append(cplex.SparsePair(ind=ind, val=[1.0] * length))
    cp.linear_constraints.add(lin_expr=lin_expr, senses=['E'] * length, rhs=[1.0] * length)

    # c.write('new.lp')

    # SOLVE THE ILP
    cp.set_results_stream(None)
    try:
        cp.solve()
    except:  # cplex.CplexSolverError:
        print("Exception raised while solving")
        return

    objective_value = cp.solution.get_objective_value()
    return objective_value


def solve_lp_matching_interval(cost_table, length_1, length_2):
    precision = length_1 * length_2
    # print(cost_table)

    c = cplex.Cplex()
    c.parameters.threads.set(1)

    # OBJECTIVE FUNCTION
    c.objective.set_sense(c.objective.sense.minimize)
    c.objective.set_name("Obj")
    objective = []
    names = []
    pos = 0
    for i in range(length_1):
        for j in range(length_2):
            names.append('x' + str(pos))
            objective.append(cost_table[i][j])
            pos += 1
    c.variables.add(obj=objective,
                    names=names,
                    types=[c.variables.type.integer] * precision)

    # FIRST GROUP OF CONSTRAINTS
    lin_expr = []
    c_names = []
    for i in range(length_1):
        ind = []
        for j in range(length_2):
            pos = i * length_2 + j
            ind.append('x' + str(pos))
        lin_expr.append(cplex.SparsePair(ind=ind, val=[1.0] * length_2))
        c_names.append('c1_' + str(i))
    c.linear_constraints.add(lin_expr=lin_expr,
                             senses=['E'] * length_1,
                             rhs=[length_2] * length_1,
                             names=c_names)

    # SECOND GROUP OF CONSTRAINTS
    lin_expr = []
    c_names = []
    for j in range(length_2):
        ind = []
        for i in range(length_1):
            pos = i * length_2 + j
            ind.append('x' + str(pos))
        lin_expr.append(cplex.SparsePair(ind=ind, val=[1.0] * length_1))
        c_names.append('c2_' + str(j))
    c.linear_constraints.add(lin_expr=lin_expr,
                             senses=['E'] * length_2,
                             rhs=[length_1] * length_2,
                             names=c_names)

    c.write('interval.lp')
    c.write('interval.mps')

    # SOLVE THE ILP
    c.set_results_stream(None)
    try:
        c.solve()
    except:  # cplex.CplexSolverError:
        print("Exception raised while solving")
        return

    result = c.solution.get_objective_value() / precision

    return result


# DODGSON SCORE
def solve_lp_file_dodgson_score(election, N=None, e=None, D=None):

    cp = cplex.Cplex()

    # OBJECTIVE FUNCTION
    cp.objective.set_sense(cp.objective.sense.minimize)

    names = []
    obj = []
    for i in range(len(N)):
        for j in range(1, len(D)):
            names.append("y" + str(i) + "_" + str(j))
            obj.append(j)
    types = [cp.variables.type.integer] * len(names)
    cp.variables.add(obj=obj, names=names, types=types)

    # ADD MISSING VARIABLES
    names = []
    for i in range(len(N)):
        names.append("y" + str(i) + "_" + str(0))
    cp.variables.add(names=list(names),
                     types=[cp.variables.type.integer] * len(names))

    # FIRST GROUP OF CONSTRAINTS
    lin_expr = []
    rhs = []
    for i in range(len(N)):
        ind = ["y" + str(i) + "_" + str(0)]
        val = [1.]
        lin_expr.append(cplex.SparsePair(ind=ind, val=val))
        rhs.append(N[i])
    cp.linear_constraints.add(lin_expr=lin_expr,
                              senses=['E'] * len(rhs),
                              rhs=rhs,
                              names=['C1_' + str(i) for i in range(len(rhs))])

    # SECOND GROUP OF CONSTRAINTS
    lin_expr = []
    rhs = []
    for k in range(len(D)):

        # if election.election_id == 'Impartial Culture_2' and k==0:
        #     continue

        val = []
        ind = []
        for i in range(len(N)):
            for j in range(1, len(D)):
                ind.append("y" + str(i) + "_" + str(j))
                value = e[i][j][k] - e[i][j - 1][k]
                val.append(value)
        lin_expr.append(cplex.SparsePair(ind=ind, val=val))
        rhs.append(D[k])

    cp.linear_constraints.add(lin_expr=lin_expr,
                              senses=['G'] * len(rhs),
                              rhs=rhs,
                              names=['C2_' + str(i) for i in range(len(rhs))])

    # THIRD GROUP OF CONSTRAINTS
    lin_expr = []
    rhs = []
    for i in range(len(N)):
        for j in range(1, len(D)):
            ind = []
            val = []
            ind.append("y" + str(i) + "_" + str(j - 1))
            ind.append("y" + str(i) + "_" + str(j))
            val.append(1.)
            val.append(-1.)
            lin_expr.append(cplex.SparsePair(ind=ind, val=val))
            rhs.append(0.)
    cp.linear_constraints.add(lin_expr=lin_expr,
                              senses=['G'] * len(rhs),
                              rhs=rhs,
                              names=['C3_' + str(i) for i in range(len(rhs))])
    # cp.write("tmp.lp")
    #
    # cp.read("tmp.lp")

    # SOLVE THE ILP
    cp.parameters.threads.set(1)
    cp.set_results_stream(None)
    try:
        cp.solve()
    except:  # cplex.CplexSolverError:
        print("Exception raised while solving")
        return

    # for x in cp.solution.get_values():
    #     if x <0 or x > 1:
    #         print(x)

    return cp.solution.get_objective_value()



def solve_lp_dodgson_score(lp_file_name):
    """ this function ..."""

    cp_lp = cplex.Cplex(lp_file_name)
    cp_lp.parameters.threads.set(1)
    cp_lp.set_results_stream(None)

    start = cp_lp.get_time()
    try:
        cp_lp.solve()
    except:  # cplex.CplexSolverError:
        print("Exception raised during solve")
        return
    stop = cp_lp.get_time()

    """
    import numpy as np
    result = np.zeros([len(N), len(D)])
    for i in range(len(N)):
        for j in range(len(D)):
            result[i] = cp_lp.solution.get_values('y' + str(i) + '_' + str(j))
    """

    return cp_lp.solution.get_objective_value(), stop-start


# FOR WINNERS - updated
def solve_lp_borda_owa(params, votes, owa):

    num_voters = params['voters']
    num_candidates = params['candidates']
    num_orders = params['orders']

    cp = cplex.Cplex()
    cp.parameters.threads.set(1)

    # OBJECTIVE FUNCTION
    cp.objective.set_sense(cp.objective.sense.maximize)
    objective = []
    names = []
    pos = 0
    for i in range(num_voters):
        for j in range(num_orders):
            for k in range(num_candidates):
                names.append(f'x{pos}')
                objective.append(owa[j])
                pos += 1
    cp.variables.add(obj=objective, names=names,
                     types=[cp.variables.type.binary] * num_voters * num_orders * num_candidates)

    # ADD VARIABLES FOR CANDIDATES
    names = [f'y{i}' for i in range(num_candidates)]
    cp.variables.add(names=list(names),
                     types=[cp.variables.type.binary] * num_candidates)

    # FIRST GROUP OF CONSTRAINTS
    lin_expr = []
    ind = [f'y{i}' for i in range(num_candidates)]
    val = [1.] * num_candidates
    lin_expr.append(cplex.SparsePair(ind=ind, val=val))
    cp.linear_constraints.add(lin_expr=lin_expr,
                              senses=['E'],
                              rhs=[num_orders])

    # SECOND GROUP OF CONSTRAINTS
    lin_expr = []
    for i in range(num_voters):
        for j in range(num_candidates):
            ind = []
            val = []
            pos = i * num_orders * num_candidates + j
            for k in range(num_orders):
                ind.append(f'x{pos + num_candidates * k}')
                val.append(1.)
            for k in range(0, j + 1):
                ind.append(f'y{int(votes[i][k])}')
                val.append(-1.)
            lin_expr.append(cplex.SparsePair(ind=ind, val=val))
    cp.linear_constraints.add(lin_expr=lin_expr,
                              senses=['L'] * num_voters * num_candidates,
                              rhs=[0.0] * num_voters * num_candidates)

    # SOLVE THE ILP
    cp.set_results_stream(None)
    start = cp.get_time()
    try:
        cp.solve()
    except cplex.CplexSolverError:
        print("Exception raised during solve")
        return
    stop = cp.get_time()

    # EXTRACT WINNERS
    result = [0.] * num_candidates
    for i in range(num_candidates):
        result[i] = cp.solution.get_values(f'y{i}')
    winner_id = 0
    winners = [0.] * num_orders
    for i in range(num_candidates):
        if math.isclose(result[i], 1.):
            winners[winner_id] = i
            winner_id += 1
    winners = sorted(winners)

    return winners, stop-start


# FOR WINNERS - needs update
def solve_lp_bloc_owa(params, votes, owa, t_bloc):
    """ this function generates lp file"""


    rand_name = str(np.random.random())
    lp_file_name = str(rand_name + ".lp")

    lp_file = open(lp_file_name, 'w')
    lp_file.write("Maximize\nobj: ")
    pos = 0
    first = True
    for i in range(params['voters']):
        for j in range(params['orders']):
            for k in range(params['candidates']):

                if not first:
                    if k == t_bloc - 1:
                        lp_file.write(" + ")
                first = False
                if k == t_bloc - 1:
                    lp_file.write(str(owa[j]) + " x" + str(pos))
                pos += 1
    lp_file.write("\n")

    lp_file.write("Subject To\n")
    lp_file.write("c0:")
    first = True
    for i in range(params['candidates']):
        if not first:
            lp_file.write(" +")
        first = False
        lp_file.write(" y" + str(i))
    lp_file.write(' = ' + str(params['orders']) + '\n')

    for i in range(params['voters']):
        for j in range(params['candidates']):
            lp_file.write("c" + str(i * params['candidates'] + j + 1) + ": ")
            pos = i * params['orders'] * params['candidates'] + j
            first = True
            for k in range(params['orders']):
                if not first:
                    lp_file.write(" +")
                first = False
                lp_file.write(" x" + str(pos + params['candidates'] * k))

            for k in range(0, j + 1):
                lp_file.write(" - y" + str(int(votes[i][k])))

            lp_file.write(" <= 0 \n")

    lp_file.write("Binary\n")
    for i in range(params['voters'] * params['orders'] * params['candidates']):
        lp_file.write("x" + str(i) + "\n")
    for i in range(params['candidates']):
        lp_file.write("y" + str(i) + "\n")

    lp_file.write("End\n")
    lp_file.close()


    cp_lp = cplex.Cplex(lp_file_name)
    cp_lp.parameters.threads.set(1)
    cp_lp.set_results_stream(None)

    start = cp_lp.get_time()
    try:
        cp_lp.solve()
    except cplex.CplexSolverError:
        print("Exception raised during solve")
        return
    stop = cp_lp.get_time()

    result = [0.] * params['candidates']
    for i in range(params['candidates']):
        result[i] = cp_lp.solution.get_values('y' + str(i))

    params['pure'] = True

    winner_id = 0
    winners = [0.] * params['orders']
    for i in range(params['candidates']):
        if math.isclose(result[i], 1.):
            if params['pure']:
                winners[winner_id] = i
            winner_id += 1
    winners = sorted(winners)

    os.remove(lp_file_name)

    return winners, stop-start


def solve_lp_matrix(lp_file_name):
    cp_lp = cplex.Cplex(lp_file_name)
    cp_lp.set_results_stream(None)

    # cp_lp.parameters.threads.set(1)
    # cp_lp.parameters.mip.tolerances.mipgap = 0.0001
    # cp_lp.parameters.mip.strategy.probe.set(3)

    try:
        cp_lp.solve()

    except:
        print("Exception raised during solve")
        return

    return cp_lp.solution.get_objective_value()


# SPEARMAN - old
def generate_ilp_distance(lp_file_name, votes_1, votes_2, params, metric_name):
    lp_file = open(lp_file_name, 'w')
    lp_file.write("Minimize\n")  # obj: ")
    #
    # first = True
    # for k in range(params['voters']):
    #     for l in range(params['voters']):
    #
    #         vote_1 = votes_1[k]
    #         vote_2 = votes_2[l]
    #
    #         if metric_name == 'spearman':
    #             pote_1 = [0] * params['candidates']
    #             pote_2 = [0] * params['candidates']
    #
    #             for i in range(params['candidates']):
    #                 pote_1[vote_1[i]] = i
    #                 pote_2[vote_2[i]] = i
    #
    #         for i in range(params['candidates']):
    #             for j in range(params['candidates']):
    #                 if not first:
    #                     lp_file.write(" + ")
    #                 first = False
    #
    #                 if metric_name == "spearman":
    #                     weight = abs(pote_1[i] - pote_2[j])
    #                 elif metric_name == "alt":
    #                     weight = float(abs(pote_1[i] - pote_2[j]) ** (2)) / float(
    #                         1. + min(pote_1[i], pote_2[j]))
    #                 elif metric_name == 'hamming':
    #                     weight = hamming(vote_1, vote_2)
    #                 else:
    #                     weight = 0
    #
    #                 lp_file.write(
    #                     str(weight) + " P" + "k" + str(k) + "l" + str(l) + "i" + str(i) + "j" + str(
    #                         j))
    # lp_file.write("\n")
    #
    # lp_file.write("Subject To\n")
    #
    # for k in range(params['voters']):
    #     for l in range(params['voters']):
    #         for i in range(params['candidates']):
    #             for j in range(params['candidates']):
    #                 lp_file.write("P" + "k" + str(k) + "l" + str(l) + "i" + str(i) + "j" + str(j))
    #                 lp_file.write(" - " + "M" + "i" + str(i) + "j" + str(j) + " <= 0" + "\n")
    #
    #                 lp_file.write("P" + "k" + str(k) + "l" + str(l) + "i" + str(i) + "j" + str(j))
    #                 lp_file.write(" - " + "N" + "k" + str(k) + "l" + str(l) + " <= 0" + "\n")
    #
    # for k in range(params['voters']):
    #     first = True
    #     for l in range(params['voters']):
    #         if not first:
    #             lp_file.write(" + ")
    #         first = False
    #         lp_file.write("N" + "k" + str(k) + "l" + str(l))
    #     lp_file.write(" = 1" + "\n")
    #
    # for l in range(params['voters']):
    #     first = True
    #     for k in range(params['voters']):
    #         if not first:
    #             lp_file.write(" + ")
    #         first = False
    #         lp_file.write("N" + "k" + str(k) + "l" + str(l))
    #     lp_file.write(" = 1" + "\n")
    #
    # for i in range(params['candidates']):
    #     first = True
    #     for j in range(params['candidates']):
    #         if not first:
    #             lp_file.write(" + ")
    #         first = False
    #         lp_file.write("M" + "i" + str(i) + "j" + str(j))
    #     lp_file.write(" = 1" + "\n")
    #
    # for j in range(params['candidates']):
    #     first = True
    #     for i in range(params['candidates']):
    #         if not first:
    #             lp_file.write(" + ")
    #         first = False
    #         lp_file.write("M" + "i" + str(i) + "j" + str(j))
    #     lp_file.write(" = 1" + "\n")
    #
    # # IMPORTANT #
    # for k in range(params['voters']):
    #     for i in range(params['candidates']):
    #         first = True
    #         for l in range(params['voters']):
    #             for j in range(params['candidates']):
    #                 if not first:
    #                     lp_file.write(" + ")
    #                 first = False
    #                 lp_file.write("P" + "k" + str(k) + "l" + str(l) + "i" + str(i) + "j" + str(j))
    #         lp_file.write(" = 1" + "\n")
    #
    # # IMPORTANT #
    # for l in range(params['voters']):
    #     for j in range(params['candidates']):
    #         first = True
    #         for k in range(params['voters']):
    #             for i in range(params['candidates']):
    #                 if not first:
    #                     lp_file.write(" + ")
    #                 first = False
    #                 lp_file.write("P" + "k" + str(k) + "l" + str(l) + "i" + str(i) + "j" + str(j))
    #         lp_file.write(" = 1" + "\n")
    #
    # lp_file.write("Binary\n")
    #
    # for k in range(params['voters']):
    #     for l in range(params['voters']):
    #         for i in range(params['candidates']):
    #             for j in range(params['candidates']):
    #                 lp_file.write(
    #                     "P" + "k" + str(k) + "l" + str(l) + "i" + str(i) + "j" + str(j) + "\n")
    #
    # for k in range(params['voters']):
    #     for l in range(params['voters']):
    #         lp_file.write("N" + "k" + str(k) + "l" + str(l) + "\n")
    #
    # for i in range(params['candidates']):
    #     for j in range(params['candidates']):
    #         lp_file.write("M" + "i" + str(i) + "j" + str(j) + "\n")

    lp_file.write("End\n")


def solve_ilp_distance(lp_file_name, votes_1, votes_2, params, metric_name):
    cp_lp = cplex.Cplex(lp_file_name)
    cp_lp.set_results_stream(None)

    # cp_lp.parameters.threads.set(1)
    # cp_lp.parameters.timelimit.set(60)

    try:
        cp_lp.solve()

    except cplex.CplexSolverError:
        print("Exception raised during solve")
        return

    total = cp_lp.solution.get_objective_value()
    return total


def spearman_cost(single_votes_1, single_votes_2, params, perm):
    pote_1 = [0] * params['candidates']
    pote_2 = [0] * params['candidates']

    for i in range(params['candidates']):
        id_1 = int(perm[0][single_votes_1[i]])
        pote_1[id_1] = i
        id_2 = int(perm[1][single_votes_2[i]])
        pote_2[id_2] = i

    total_diff = 0.
    for i in range(params['candidates']):
        local_diff = float(abs(pote_1[i] - pote_2[i]))
        total_diff += local_diff

    return total_diff


def spearman_cost_per_cand(single_votes_1, single_votes_2, params, perm):
    pote_1 = [0] * params['candidates']
    pote_2 = [0] * params['candidates']

    for i in range(params['candidates']):
        id_1 = int(perm[0][single_votes_1[i]])
        pote_1[id_1] = i
        id_2 = int(perm[1][single_votes_2[i]])
        pote_2[id_2] = i

    cand_diff = [0] * params['candidates']
    for i in range(params['candidates']):
        cand_diff[i] = float(abs(pote_1[i] - pote_2[i]))

    return cand_diff


def remove_lp_file(path):
    """ Safely remove lp file """
    with suppress(OSError):
        os.remove(path)


# COPY
def generate_lp_file_borda_owa(owa, lp_file_name, params, votes):
    """ this function generates lp file"""

    with open(lp_file_name, 'w') as lp_file:
        lp_file.write("Maximize\nobj: ")
        pos = 0
        first = True
        for i in range(params['voters']):
            for j in range(params['orders']):
                for k in range(params['candidates']):
                    if not first and owa[j] >= 0.:
                        lp_file.write(" + ")
                    first = False
                    lp_file.write(str(owa[j]) + " x" + str(pos))
                    pos += 1
        lp_file.write("\n")

        lp_file.write("Subject To\n")
        lp_file.write("c0:")
        first = True
        for i in range(params['candidates']):
            if not first:
                lp_file.write(" +")
            first = False
            lp_file.write(" y" + str(i))
        lp_file.write(' = ' + str(params['orders']) + '\n')

        for i in range(params['voters']):
            for j in range(params['candidates']):
                lp_file.write("c" + str(i * params['candidates'] + j + 1) + ": ")
                pos = i * params['orders'] * params['candidates'] + j
                first = True
                for k in range(params['orders']):
                    if not first:
                        lp_file.write(" +")
                    first = False
                    lp_file.write(" x" + str(pos + params['candidates'] * k))

                for k in range(0, j + 1):
                    lp_file.write(" - y" + str(int(votes[i][k])))

                lp_file.write(" <= 0 \n")

        lp_file.write("Binary\n")
        for i in range(params['voters'] * params['orders'] * params['candidates']):
            lp_file.write("x" + str(i) + "\n")
        for i in range(params['candidates']):
            lp_file.write("y" + str(i) + "\n")

        lp_file.write("End\n")


# TMP - PRINT

def print_results(election, committee_size, cp):

    result = [0.] * election.num_candidates
    for i in range(election.num_candidates):
        result[i] = cp.solution.get_values('y' + str(i))
    # print(result)

    fixed = []
    for i, r in enumerate(result):
        if r >= 0.5:
            fixed.append(i)


    winner_id = 0
    winners = [-1] * committee_size
    for i in range(election.num_candidates):
        if result[i] == 1.:
            winners[winner_id] = i
            winner_id += 1
    winners = sorted(winners)

    def get_pav_score(election, winners) -> float:

        num_voters = election.num_voters
        num_candidates = election.num_candidates
        votes = election.votes

        score = 0

        vector = [0.] * num_candidates
        for i in range(len(winners)):
            vector[i] = 1.

        for i in range(num_voters):
            ctr = 1.
            for j in range(num_candidates):
                if votes[i][j] in winners:
                    score += (1. / ctr) * vector[j]
                    ctr += 1

        return score

    try:
        score = get_pav_score(election, winners)
    except:
        score = -1

    print(score)
    return score, fixed


import ast
import csv
import os
import re
from collections import Counter

import numpy as np

from mapel.core.glossary import *

regex_file_name = r'# FILE NAME:'
regex_title = r'# TITLE:'
regex_data_type = r'# DATA TYPE:'
regex_number_alternatives = r"# NUMBER ALTERNATIVES:"
regex_number_voters = r"# NUMBER VOTERS:"
regex_number_unique_orders = r"# NUMBER UNIQUE ORDERS:"
regex_number_categories = r"# NUMBER CATEGORIES:"
regex_culture_id = r"# CULTURE ID:"
regex_params = r"# PARAMS:"


def import_distances(experiment,
                     object_type: str = 'vote'):
    """
    Imports distances from a csv file.

    Parameters
    ----------
        experiment
            Experiment.
        object_type : str
            Object type.
    """

    file_name = f'{experiment.election_id}_{object_type}.csv'
    path = os.path.join(os.getcwd(), 'experiments', experiment.experiment_id, 'distances',
                        file_name)

    with open(path, 'r', newline='') as csv_file:
        reader = csv.DictReader(csv_file, delimiter=';')
        length = int(len(list(reader)) ** 0.5)

    with open(path, 'r', newline='') as csv_file:
        reader = csv.DictReader(csv_file, delimiter=';')
        distances = np.zeros([length, length])
        for row in reader:
            distances[int(row['v1'])][int(row['v2'])] = float(row['distance'])
            distances[int(row['v2'])][int(row['v1'])] = float(row['distance'])

    return distances


def import_coordinates(experiment,
                       object_type: str = 'vote'):
    """
    Imports coordinates from a csv file.

    Parameters
    ----------
        experiment
            Experiment.
        object_type : str
            Object type.
    """

    file_name = f'{experiment.election_id}_{object_type}.csv'
    path = os.path.join(os.getcwd(), 'experiments', experiment.experiment_id, 'coordinates',
                        file_name)

    with open(path, 'r', newline='') as csv_file:
        reader = csv.DictReader(csv_file, delimiter=';')
        length = len(list(reader))

    with open(path, 'r', newline='') as csv_file:
        reader = csv.DictReader(csv_file, delimiter=';')
        coordinates = np.zeros([length, 2])
        for row in reader:
            coordinates[int(row['vote_id'])] = [float(row['x']), float(row['y'])]

    return coordinates


def process_soc_line(line: str, votes: list):
    tokens = line.split(':')
    nr_this_vote = int(tokens[0])
    vote = [int(x) for x in tokens[1].split(',')]
    vote = np.array(vote)
    for i in range(0, nr_this_vote):
        votes.append(vote)
    pass


def process_app_line(line: str, votes: list):
    tokens = line.split(':')
    nr_this_vote = int(tokens[0])
    vote = set(ast.literal_eval(" ".join(tokens[1])))
    for i in range(0, nr_this_vote):
        votes.append(vote)
    pass


def process_soi_line(line: str, votes: list):
    pass


def process_toc_line(line: str, votes: list):
    pass


def process_toi_line(line: str, votes: list):
    pass


def import_real_new_soc_election(experiment_id: str = None,
                                 election_id: str = None,
                                 is_shifted=False,
                                 file_ending=4):
    """ Import real ordinal election form .soc file """

    file_name = f'{election_id}.soc'
    path = os.path.join(os.getcwd(), "experiments", experiment_id, "elections", file_name)
    file = open(path, 'r')

    params = None
    culture_id = None
    votes = []
    num_candidates = 0
    nr_votes = 0
    nr_unique = 0
    alternative_names = list()
    from_file_file_name = ''
    from_file_title = ''
    from_file_data_type = ''
    # read metadata
    for line in file:
        if line[-1] == '\n':
            line = line[:-1]
        if line[0] != '#':
            break
        elif re.search(regex_file_name, line):
            from_file_file_name = line.split(':')[1][1:-file_ending]
        elif re.search(regex_title, line):
            from_file_title = line.split(':')[1].replace(" ", "")
        elif re.search(regex_data_type, line):
            from_file_data_type = line.split(':')[1].replace(" ", "")
        elif re.search(regex_number_alternatives, line):
            num_candidates = int(line.split(':')[1])
        elif re.search(regex_number_voters, line):
            num_voters = int(line.split(':')[1])
        elif re.search(regex_number_unique_orders, line):
            nr_unique = int(line.split(':')[1])
        elif re.search(regex_culture_id, line):
            culture_id = str(line.split(':')[1])
        elif re.search(regex_params, line):
            line = line.strip().split()
            if len(line) <= 2:
                params = {}
            else:
                params = ast.literal_eval(" ".join(line[2:]))

    # label = from_file_title + "_" + from_file_file_name
    # read votes
    if from_file_data_type == 'soc':
        for line in file:
            process_soc_line(line, votes)
    elif from_file_data_type == 'soi':
        for line in file:
            process_soi_line(line, votes)
    elif from_file_data_type == 'toc':
        for line in file:
            process_toc_line(line, votes)
    elif from_file_data_type == 'toi':
        for line in file:
            process_toi_line(line, votes)
    else:
        print("Unknown data format.")

    file.close()

    alliances = None

    c = Counter(map(tuple, votes))
    counted_votes = [[count, list(row)] for row, count in c.items()]
    counted_votes = sorted(counted_votes, reverse=True)
    quantites = [a[0] for a in counted_votes]
    distinct_votes = [a[1] for a in counted_votes]
    num_options = len(counted_votes)

    if is_shifted:
        votes = [[vote - 1 for vote in voter] for voter in votes]

    return np.array(votes), \
           len(votes), \
           num_candidates, \
           params, \
           culture_id, \
           alliances, \
           num_options, \
           quantites, \
           distinct_votes


def import_real_old_soc_election(experiment_id: str = None,
                                 election_id: str = None,
                                 is_shifted=False):
    """ Import real ordinal election form .soc file """

    file_name = f'{election_id}.soc'
    path = os.path.join(os.getcwd(), "experiments", experiment_id, "elections", file_name)
    my_file = open(path, 'r')

    params = 0
    first_line = my_file.readline()

    if first_line[0] != '#':
        culture_id = 'empty'
        num_candidates = int(first_line)
    else:
        first_line = first_line.strip().split()
        culture_id = first_line[1]
        if experiment_id == 'original_ordinal_map':
            params = {}
            culture_id = _old_name_extractor(first_line)
        else:
            if len(first_line) <= 2:
                params = {}
            else:
                params = ast.literal_eval(" ".join(first_line[2:]))

        num_candidates = int(my_file.readline())

    alliances = {}
    for i in range(num_candidates):
        line = my_file.readline().strip().split()
        if len(line) > 2:
            alliances[i] = int(line[2])

    line = my_file.readline().rstrip("\n").split(',')
    num_voters = int(line[0])
    num_options = int(line[2])
    votes = [[0 for _ in range(num_candidates)] for _ in range(num_voters)]

    it = 0
    quantites = []
    for j in range(num_options):
        line = list(map(int, my_file.readline().rstrip("\n").split(',')))
        quantity = line[0]
        quantites.append(quantity)

        for k in range(quantity):
            votes[it] = line[1:num_candidates + 1]
            it += 1

    c = Counter(map(tuple, votes))
    counted_votes = [[count, list(row)] for row, count in c.items()]
    counted_votes = sorted(counted_votes, reverse=True)
    quantites = [a[0] for a in counted_votes]
    distinct_votes = [a[1] for a in counted_votes]
    num_options = len(counted_votes)

    if is_shifted:
        votes = [[vote - 1 for vote in voter] for voter in votes]
    my_file.close()

    return np.array(votes), \
           num_voters, \
           num_candidates, \
           params, \
           culture_id, \
           alliances, \
           num_options, \
           quantites, \
           distinct_votes


def import_real_soc_election(**kwargs):
    try:
        return import_real_old_soc_election(**kwargs)
    except:
        return import_real_new_soc_election(**kwargs)


def import_fake_soc_election(experiment_id, name):
    """ Import fake ordinal election form .soc file """

    file_name = f'{name}.soc'
    path = os.path.join(os.getcwd(), "experiments", experiment_id, "elections", file_name)
    my_file = open(path, 'r')

    first_line = my_file.readline()
    first_line = first_line.strip().split()
    culture_id = first_line[1]
    if len(first_line) <= 2:
        params = {}
    else:
        params = ast.literal_eval(" ".join(first_line[2:]))

    num_candidates = int(my_file.readline())
    num_voters = int(my_file.readline())

    my_file.close()

    return culture_id, params, num_voters, num_candidates


def import_real_new_app_election(experiment_id: str = None,
                                 election_id: str = None,
                                 is_shifted: bool = False,
                                 file_ending=4):
    """ Import real approval election form .app file """

    file_name = f'{election_id}.app'
    path = os.path.join(os.getcwd(), "experiments", experiment_id, "elections", file_name)
    file = open(path, 'r')

    params = None
    culture_id = None
    votes = []
    num_candidates = 0
    nr_votes = 0
    nr_unique = 0
    alternative_names = list()
    from_file_file_name = ''
    from_file_title = ''
    from_file_data_type = ''
    # read metadata
    for line in file:
        if line[-1] == '\n':
            line = line[:-1]
        if line[0] != '#':
            break
        elif re.search(regex_file_name, line):
            from_file_file_name = line.split(':')[1][1:-file_ending]
        elif re.search(regex_title, line):
            from_file_title = line.split(':')[1].replace(" ", "")
        elif re.search(regex_data_type, line):
            from_file_data_type = line.split(':')[1].replace(" ", "")
        elif re.search(regex_number_alternatives, line):
            num_candidates = int(line.split(':')[1])
        elif re.search(regex_number_voters, line):
            num_voters = int(line.split(':')[1])
        elif re.search(regex_number_unique_orders, line):
            nr_unique = int(line.split(':')[1])
        elif re.search(regex_culture_id, line):
            culture_id = str(line.split(':')[1])
        elif re.search(regex_params, line):
            line = line.strip().split()

            if len(line) <= 2:
                params = {}
            else:
                params = ast.literal_eval(" ".join(line[2:]))

    # label = from_file_title + "_" + from_file_file_name
    # read votes
    if from_file_data_type == 'app':
        for line in file:
            process_app_line(line, votes)
    else:
        print("Unknown data format.")

    file.close()

    c = Counter(map(tuple, votes))
    counted_votes = [[count, list(row)] for row, count in c.items()]
    counted_votes = sorted(counted_votes, reverse=True)
    quantites = [a[0] for a in counted_votes]
    distinct_votes = [a[1] for a in counted_votes]
    num_options = len(counted_votes)

    if is_shifted:
        votes = [[vote - 1 for vote in voter] for voter in votes]

    num_voters = len(votes)

    return votes, \
           num_voters, \
           num_candidates, \
           params, \
           culture_id, \
           num_options, \
           quantites, \
           distinct_votes


def import_real_old_app_election(experiment_id: str, election_id: str, is_shifted=False):
    """ Import real approval election from .app file """

    file_name = f'{election_id}.app'
    path = os.path.join(os.getcwd(), "experiments", experiment_id, "elections", file_name)
    my_file = open(path, 'r')

    params = 0
    first_line = my_file.readline()
    if first_line[0] != '#':
        culture_id = 'empty'
        num_candidates = int(first_line)
    else:
        first_line = first_line.strip().split()
        culture_id = first_line[1]
        if len(first_line) <= 2:
            params = {}
        else:
            params = ast.literal_eval(" ".join(first_line[2:]))

        num_candidates = int(my_file.readline())

    for _ in range(num_candidates):
        my_file.readline()

    line = my_file.readline().rstrip("\n").split(',')
    num_voters = int(line[0])
    num_options = int(line[2])
    votes = [set() for _ in range(num_voters)]

    it = 0
    for j in range(num_options):
        line = my_file.readline().rstrip("\n").replace("{", ''). \
            replace("}", '').replace(' ', '').split(',')
        if line[1] != '':
            quantity = int(line[0])
            for k in range(quantity):
                for el in range(len(line) - 1):
                    votes[it].add(int(line[el + 1]))
                it += 1

    if culture_id in NICE_NAME.values():
        rev_dict = dict(zip(NICE_NAME.values(), NICE_NAME.keys()))
        culture_id = rev_dict[culture_id]

    if is_shifted:
        votes = [{c - 1 for c in vote} for vote in votes]
    my_file.close()

    c = Counter(map(tuple, votes))
    counted_votes = [[count, list(row)] for row, count in c.items()]
    counted_votes = sorted(counted_votes, reverse=True)
    quantites = [a[0] for a in counted_votes]
    distinct_votes = [a[1] for a in counted_votes]
    num_options = len(counted_votes)

    return votes, \
           num_voters, \
           num_candidates, \
           params, \
           culture_id, \
           num_options, \
           quantites, \
           distinct_votes


def import_real_app_election(**kwargs):
    try:
        return import_real_old_app_election(**kwargs)
    except:
        return import_real_new_app_election(**kwargs)


def import_fake_app_election(experiment_id: str, name: str):
    """ Import fake approval election from .app file """

    file_name = f'{name}.app'
    path = os.path.join(os.getcwd(), "experiments", experiment_id, "elections", file_name)
    my_file = open(path, 'r')
    first_line = my_file.readline()
    first_line = first_line.strip().split()
    fake_culture_id = first_line[1]
    if len(first_line) <= 2:
        params = {}
    else:
        params = ast.literal_eval(" ".join(first_line[2:]))

    num_candidates = int(my_file.readline().strip())
    num_voters = int(my_file.readline().strip())

    return fake_culture_id, params, num_voters, num_candidates


def check_if_fake(experiment_id, name, extention):
    file_name = f'{name}.{extention}'
    path = os.path.join(os.getcwd(), "experiments", experiment_id, "elections", file_name)
    my_file = open(path, 'r')
    line = my_file.readline().strip()
    my_file.close()
    return line[0] == '$'


def _old_name_extractor(first_line):
    if len(first_line) == 4:
        culture_id = f'{first_line[1]} {first_line[2]} {first_line[3]}'
    elif len(first_line) == 3:
        culture_id = f'{first_line[1]} {first_line[2]}'
    elif len(first_line) == 2:
        culture_id = first_line[1]
    else:
        culture_id = 'noname'
    return culture_id



#!/usr/bin/env python

import logging
from typing import Union

from mapel.core.glossary import *

import mapel.elections.cultures.to_be_removed.mallows_urn as mallows_urn

import mapel.elections.cultures.guardians as guardians
import mapel.elections.cultures.guardians_plus as guardians_plus
import mapel.elections.cultures.mallows as mallows
import mapel.elections.cultures.matrices.single_peaked_matrices as sp_matrices
import mapel.elections.cultures.unused as unused

from mapel.elections.cultures.nonstandard.alliances import *
import mapel.elections.cultures.nonstandard.field_experiment as fe
from mapel.elections.cultures.preflib import generate_preflib_votes

import mapel.elections.cultures.group_separable as group_separable
import mapel.elections.cultures.euclidean as euclidean
import mapel.elections.cultures.urn as urn

import prefsampling.ordinal as pref_ordinal
import prefsampling.approval as pref_approval

registered_approval_cultures = {
    'impartial': pref_approval.impartial,
    'impartial_culture': pref_approval.impartial,
    'ic': pref_approval.impartial,
    'id': pref_approval.identity,
    'resampling': pref_approval.resampling,
    'disjoint_resampling': pref_approval.disjoint_resampling,
    'moving_resampling': pref_approval.moving_resampling,
    'noise': pref_approval.noise,
    'euclidean': pref_approval.euclidean,
    'full': pref_approval.full,
    'empty': pref_approval.empty,
    'truncated_urn': urn.truncated_urn_mask,
    'urn_partylist': pref_approval.urn_partylist,

    'field': fe.generate_approval_field_votes,  # unsupported culture
    'truncated_mallows': mallows.generate_approval_truncated_mallows_votes,  # unsupported culture

    'approval_full': pref_approval.full,  # deprecated name
    'approval_empty': pref_approval.empty,  # deprecated name
}

registered_ordinal_cultures = {
    'id': pref_ordinal.identity,
    'impartial': pref_ordinal.impartial,
    'impartial_culture': pref_ordinal.impartial,
    'ic': pref_ordinal.impartial,
    'iac': pref_ordinal.impartial_anonymous,

    'urn': pref_ordinal.urn,
    'single-crossing': pref_ordinal.single_crossing,
    'conitzer': pref_ordinal.single_peaked_conitzer,
    'walsh': pref_ordinal.single_peaked_walsh,
    'spoc': pref_ordinal.single_peaked_circle,
    'stratification': pref_ordinal.stratification,
    'mallows': pref_ordinal.mallows,
    'didi': pref_ordinal.didi,

    'group-separable': group_separable.gs_mask,
    'euclidean': euclidean.euclidean_mask,

    'norm-mallows': mallows.generate_mallows_votes,
    'real_identity': guardians.generate_real_identity_votes,

    'plackett-luce': pref_ordinal.plackett_luce,

    'mallows_urn': mallows_urn.generate_mallows_urn_votes,
    'idan_part': guardians_plus.generate_idan_part_votes,  # unsupported culture
    'idun_part': guardians_plus.generate_idun_part_votes,  # unsupported culture
    'idst_part': guardians_plus.generate_idst_part_votes,  # unsupported culture
    'anun_part': guardians_plus.generate_anun_part_votes,  # unsupported culture
    'anst_part': guardians_plus.generate_anst_part_votes,  # unsupported culture
    'unst_part': guardians_plus.generate_unst_part_votes,  # unsupported culture
    'idan_mallows': guardians_plus.generate_idan_mallows_votes,  # unsupported culture
    'idst_mallows': guardians_plus.generate_idst_mallows_votes,  # unsupported culture
    'anun_mallows': guardians_plus.generate_anun_mallows_votes,  # unsupported culture
    'unst_mallows': guardians_plus.generate_unst_mallows_votes,  # unsupported culture
    'unst_topsize': guardians_plus.generate_unst_topsize_votes,  # unsupported culture
    'idst_blocks': guardians_plus.generate_idst_blocks_votes,
    'norm-mallows_mixture': mallows.generate_norm_mallows_mixture_votes,  # unsupported culture
    'walsh_mallows': sp_matrices.generate_walsh_mallows_votes,  # unsupported culture
    'conitzer_mallows': sp_matrices.generate_conitzer_mallows_votes,  # unsupported culture
    'mallows_triangle': mallows.generate_mallows_votes,  # unsupported culture
    'walsh_party': unused.generate_sp_party,  # unsupported culture
    'conitzer_party': unused.generate_sp_party,  # unsupported culture
    'mallows_party': mallows.generate_mallows_party,  # unsupported culture
    'ic_party': unused.generate_ic_party,  # unsupported culture
    'real_uniformity': guardians.generate_real_uniformity_votes,  # unsupported culture
    'real_antagonism': guardians.generate_real_antagonism_votes,  # unsupported culture
    'real_stratification': guardians.generate_real_stratification_votes,  # unsupported culture
    'un_from_matrix': guardians_plus.generate_un_from_matrix_votes,  # unsupported culture
    'un_from_list': guardians_plus.generate_un_from_list,  # unsupported culture

    'urn_model': pref_ordinal.urn,  # deprecated name
    'weighted_stratification': pref_ordinal.stratification,  # deprecated name
}


def generate_approval_votes(
        culture_id: str = None,
        num_voters: int = None,
        num_candidates: int = None,
        params: dict = None
) -> Union[list, np.ndarray]:
    """
    Generates approval votes according to the given culture id.

    Parameters
    ----------
        culture_id : str
            Name of the culture.
        num_voters : int
            Number of Voters.
        num_candidates : int
            Number of Candidates
        params : dict
            Culture parameters.
    """
    if culture_id in registered_approval_cultures:
        return registered_approval_cultures.get(culture_id)(num_voters, num_candidates, **params)

    else:
        logging.warning(f'No such culture id: {culture_id}')
        return []


def generate_ordinal_votes(
        culture_id: str = None,
        num_candidates: int = None,
        num_voters: int = None,
        params: dict = None,
        **kwargs
) -> Union[list, np.ndarray]:
    """
    Generates approval votes according to the given culture id.

    Parameters
    ----------
        culture_id : str
            Name of the culture.
        num_voters : int
            Number of Voters.
        num_candidates : int
            Number of Candidates
        params : dict
            Culture parameters.
    """

    if culture_id in LIST_OF_PREFLIB_MODELS:
        try:
            votes = generate_preflib_votes(culture_id=culture_id,
                                           num_candidates=num_candidates,
                                           num_voters=num_voters,
                                           params=params)
        except:
            votes = []
            logging.warning(
                f'You are trying to create an election based on Preflib '
                f'without having the original source election. '
                f'Please use different culture_id than: {culture_id}')

    elif culture_id in registered_ordinal_cultures:
        votes = registered_ordinal_cultures.get(culture_id)(num_voters=num_voters,
                                                            num_candidates=num_candidates,
                                                            **params)

    elif culture_id in LIST_OF_FAKE_MODELS:
        votes = [culture_id, num_candidates, num_voters, params]

    else:
        votes = []
        logging.warning(
            f'No such culture id: {culture_id} \n'
            f'If you are using your own instances then ignore this warning.')

    if culture_id not in LIST_OF_FAKE_MODELS:
        votes = [[int(x) for x in row] for row in votes]

    return np.array(votes)


def approval_votes_to_vectors(votes, num_candidates=None, num_voters=None):
    vectors = np.zeros([num_candidates, num_candidates])
    for vote in votes:
        denom_in = len(vote)
        denom_out = num_candidates - denom_in
        for i in range(num_candidates):
            if i in vote:
                for j in range(denom_in):
                    vectors[i][j] += 1 / denom_in / num_voters
            else:
                for j in range(denom_out):
                    vectors[i][denom_in + j] += 1 / denom_out / num_voters
    return vectors


def from_approval(num_candidates: int = None,
                  num_voters: int = None,
                  params: dict = None):
    votes = generate_approval_votes(culture_id=params['culture_id'],
                                    num_candidates=num_candidates, num_voters=num_voters,
                                    params=params)

    return approval_votes_to_vectors(votes, num_candidates=num_candidates, num_voters=num_voters)


LIST_OF_ORDINAL_ALLIANCE_MODELS = {
    'ic': generate_ordinal_alliance_ic_votes,
    'urn': generate_ordinal_alliance_urn_votes,
    'euc': generate_ordinal_alliance_euclidean_votes,
    'allied_euc': generate_ordinal_alliance_allied_euclidean_votes,
    'norm-mallows': generate_ordinal_alliance_norm_mallows_votes,
}


def generate_ordinal_alliance_votes(culture_id: str = None,
                                    num_candidates: int = None,
                                    num_voters: int = None,
                                    params: dict = None):
    if culture_id in LIST_OF_ORDINAL_ALLIANCE_MODELS:
        votes, alliances = LIST_OF_ORDINAL_ALLIANCE_MODELS.get(culture_id)(
            num_voters=num_voters,
            num_candidates=num_candidates,
            params=params)
    else:
        votes = []
        alliances = []
        logging.warning(f'No such culture id: {culture_id}')

    return np.array(votes), alliances


def add_approval_culture(name, function):
    """
    Adds a new approval culture to the list of available approval cultures.

    Parameters
    ----------
        name:
            Name of the culture, which will be used as culture id.
        function : str
            Function that generates the votes.
    """
    registered_approval_cultures[name] = function


def add_ordinal_culture(name, function):
    """
    Adds a new ordinal culture to the list of available ordinal cultures.

    Parameters
    ----------
        name:
            Name of the culture, which will be used as culture id.
        function : str
            Function that generates the votes.
    """
    registered_ordinal_cultures[name] = function



#!/usr/bin/env python

from mapel.core.matchings import solve_matching_vectors


def get_matching_cost_committee(election, committee_1, committee_2, distance_id):
    if distance_id == 'discrete':
        return len(committee_1.symmetric_difference(committee_2))
    elif distance_id == 'hamming':
        return hamming_distance_between_committees(election, committee_1, committee_2)
    elif distance_id == 'asymmetric':
        return asymmetric_distance_between_committees(election, committee_1, committee_2)


# MAIN FUNCTIONS
def hamming_distance_between_committees(election, committee_1, committee_2):
    cost_table = get_matching_cost_committee_hamming(election,
                                                     list(committee_1), list(committee_2))
    return solve_matching_vectors(cost_table)[0]


def asymmetric_distance_between_committees(election, committee_1, committee_2):
    cost_table = get_matching_cost_committee_asymmetric(election,
                                                        list(committee_1), list(committee_2))
    return solve_matching_vectors(cost_table)[0]


# HELPER FUNCTIONS
def get_matching_cost_committee_hamming(election, committee_1, committee_2):
    size = len(committee_1)
    return [[election.candidatelikeness_original_vectors[committee_1[i]][committee_2[j]]
             for i in range(size)] for j in range(size)]


def get_matching_cost_committee_asymmetric(election, committee_1, committee_2):
    size = len(committee_1)
    election.compute_reverse_approvals()
    return [[compare_candidates(election, committee_1[i], committee_2[j])
             for i in range(size)] for j in range(size)]


def compare_candidates(election, c1, c2):
    app_c1 = election.reverse_approvals[c1]
    app_c2 = election.reverse_approvals[c2]
    if len(app_c1.union(app_c2)) == 0:
        return 1
    return 1 - len(app_c1.intersection(app_c2)) / len(app_c1.union(app_c2))



#!/usr/bin/env python

import math
import os
import sys

import numpy as np
try:
    import pulp
except Exception:
    pulp = None

try:
    from dotenv import load_dotenv
    load_dotenv()
    sys.path.append(os.environ["PATH"])
    from abcvoting import abcrules, preferences
except ImportError:
    abcrules = None
    preferences = None

from mapel.core.glossary import *
from mapel.elections.distances import lp
from mapel.elections.other import winners as win


# MAIN FUNCTIONS
def highest_borda_score(election) -> dict:
    """
    Computes the highest Borda score of a given election

        Parameters
        ----------
        election : OrdinalElection

        Returns
        -------
        dict
            'value': highest Borda score

    """
    if election.culture_id in LIST_OF_FAKE_MODELS:
        return {'value': None}
    c = election.num_candidates
    vectors = election.get_vectors()
    borda = [sum([vectors[i][pos] * (c - pos - 1) for pos in range(c)])
             for i in range(c)]
    return {'value': max(borda) * election.num_voters}


def highest_plurality_score(election) -> dict:
    """
    Computes the highest Plurality score of a given election

        Parameters
        ----------
        election : OrdinalElection

        Returns
        -------
        dict
            'value': highest Plurality score

    """
    if election.culture_id in LIST_OF_FAKE_MODELS:
        return {'value': None}
    first_pos = election.get_matrix()[0]
    return {'value': max(first_pos) * election.num_voters}


def highest_copeland_score(election) -> dict:
    """
    Computes the highest Copeland score of a given election

        Parameters
        ----------
        election : OrdinalElection

        Returns
        -------
        dict
            'value': highest Copeland score

    """
    if election.culture_id in LIST_OF_FAKE_MODELS:
        return {'value': None}

    election.compute_potes()
    scores = np.zeros([election.num_candidates])

    for i in range(election.num_candidates):
        for j in range(i + 1, election.num_candidates):
            result = 0
            for k in range(election.num_voters):
                if election.potes[k][i] < election.potes[k][j]:
                    result += 1
            if result > election.num_voters / 2:
                scores[i] += 1
            elif result < election.num_voters / 2:
                scores[j] += 1
            else:
                scores[i] += 0.5
                scores[j] += 0.5

    return {'value': max(scores)}


def lowest_dodgson_score(election):
    """
    Computes the lowest Dodgson score of a given election

        Parameters
        ----------
        election : OrdinalElection

        Returns
        -------
        dict
            'value': lowest Dodgson score

    """
    if election.culture_id in LIST_OF_FAKE_MODELS:
        return {'value': None}

    min_score = math.inf

    for target_id in range(election.num_candidates):

        # PREPARE N
        unique_potes, N = _potes_to_unique_potes(election.get_potes())

        e = np.zeros([len(N), election.num_candidates,
                      election.num_candidates])

        # PREPARE e
        for i, p in enumerate(unique_potes):
            for j in range(election.num_candidates):
                for k in range(election.num_candidates):
                    if p[target_id] <= p[k] + j:
                        e[i][j][k] = 1

        # PREPARE D
        D = [0 for _ in range(election.num_candidates)]
        threshold = math.ceil((election.num_voters + 1) / 2.)
        for k in range(election.num_candidates):
            diff = 0
            for i, p in enumerate(unique_potes):
                if p[target_id] < p[k]:
                    diff += N[i]
                if diff >= threshold:
                    D[k] = 0
                else:
                    D[k] = threshold - diff
        D[target_id] = 0  # always winning

        score = lp.solve_lp_file_dodgson_score(election, N=N, e=e, D=D)

        if score < min_score:
            min_score = score

    return min_score


def highest_cc_score(election, committee_size=1):
    if election.culture_id in LIST_OF_FAKE_MODELS:
        return {'value': None, 'dissat': None}
    winners, total_time = win.generate_winners(election=election,
                                             num_winners=committee_size,
                                             ballot="ordinal",
                                             type='borda_owa', name='cc')
    return {'value': get_cc_score(election, winners), 'dissat': get_cc_dissat(election, winners)}


def highest_hb_score(election, committee_size=1):
    if election.culture_id in LIST_OF_FAKE_MODELS:
        return {'value': None, 'dissat': None}
    winners, total_time = win.generate_winners(election=election,
                                             num_winners=committee_size,
                                             ballot="ordinal",
                                             type='borda_owa', name='hb')
    return {'value': get_hb_score(election, winners), 'dissat': get_hb_dissat(election, winners)}


def highest_pav_score(election, committee_size=1):
    if election.culture_id in LIST_OF_FAKE_MODELS:
        return {'value': None, 'dissat': None}
    winners, total_time = win.generate_winners(election=election,
                                             num_winners=committee_size,
                                             ballot="ordinal",
                                             type='bloc_owa', name='hb')
    return {'value': get_pav_score(election, winners), 'dissat': get_pav_dissat(election, winners)}


# HELPER FUNCTIONS
def _potes_to_unique_potes(potes):
    """ Remove repetitions from potes (positional votes) """
    unique_potes = []
    n = []
    for pote in potes:
        flag_new = True
        for i, p in enumerate(unique_potes):
            if list(pote) == list(p):
                n[i] += 1
                flag_new = False
        if flag_new:
            unique_potes.append(pote)
            n.append(1)
    return unique_potes, n


# GET SCORE
def get_score(election, winners, rule) -> float:
    if rule == 'cc':
        return get_cc_score(election, winners)
    elif rule == 'hb':
        return get_hb_score(election, winners)
    elif rule == 'pav':
        return get_pav_score(election, winners)


def get_cc_score(election, winners) -> float:
    num_voters = election.num_voters
    num_candidates = election.num_candidates
    votes = election.votes

    score = 0

    for i in range(num_voters):
        for j in range(num_candidates):
            if votes[i][j] in winners:
                score += num_candidates - j - 1
                break

    return score


def get_hb_score(election, winners) -> float:
    num_voters = election.num_voters
    num_candidates = election.num_candidates
    votes = election.votes

    score = 0

    for i in range(num_voters):
        ctr = 1.
        for j in range(num_candidates):
            if votes[i][j] in winners:
                score += (1. / ctr) * (num_candidates - j - 1)
                ctr += 1

    return score


def get_pav_score(election, winners) -> float:
    num_voters = election.num_voters
    num_candidates = election.num_candidates
    votes = election.votes

    score = 0

    vector = [0.] * num_candidates
    for i in range(len(winners)):
        vector[i] = 1.

    for i in range(num_voters):
        ctr = 1.
        for j in range(num_candidates):
            if votes[i][j] in winners:
                score += (1. / ctr) * vector[j]
                ctr += 1

    return score


# GET DISSAT
def get_dissat(election, winners, rule) -> float:
    if rule == 'cc':
        return get_cc_dissat(election, winners)
    elif rule == 'hb':
        return get_hb_dissat(election, winners)
    elif rule == 'pav':
        return get_pav_dissat(election, winners)


def get_cc_dissat(election, winners) -> float:

    num_voters = election.num_voters
    num_candidates = election.num_candidates

    dissat = 0

    for i in range(num_voters):
        for j in range(num_candidates):
            if election.votes[i][j] in winners:
                dissat += j
                break

    return dissat


def get_hb_dissat(election, winners) -> float:

    num_voters = election.num_voters
    num_candidates = election.num_candidates

    dissat = 0

    for i in range(num_voters):
        ctr = 1.
        for j in range(num_candidates):
            if election.votes[i][j] in winners:
                dissat += (1. / ctr) * (j)
                ctr += 1

    return dissat


def get_pav_dissat(election, winners) -> float:

    num_voters = election.num_voters
    num_candidates = election.num_candidates

    dissat = 0

    vector = [0. for _ in range(num_candidates)]
    for i in range(len(winners), num_candidates):
        vector[i] = 1.

    for i in range(num_voters):
        ctr = 1.
        for j in range(num_candidates):
            if election.votes[i][j] in winners:
                dissat += ((1. / ctr) * vector[j])
                ctr += 1

    return dissat


# OTHER
def borda_spread(election) -> int:
    """ Compute the difference between the highest and the lowest Borda score """
    c = election.num_candidates
    vectors = election.get_vectors()
    borda = [sum([vectors[i][pos] * (c - pos - 1) for pos in range(c)])
             for i in range(c)]
    return (max(borda)-min(borda)) * election.num_voters



import numpy as np


def convert_ordinal_to_approval(votes):
    approval_votes = [{} for _ in range(len(votes))]
    for i, vote in enumerate(votes):
        k = int(np.random.beta(2, 6) * len(votes[0])) + 1
        approval_votes[i] = set(vote[0:k])

    return approval_votes


#!/usr/bin/env python

from typing import Callable

import os

from mapel.core.matchings import *
from mapel.elections.objects.ApprovalElection import ApprovalElection


# MAIN APPROVAL DISTANCES

def compute_approvalwise(election_1: ApprovalElection, election_2: ApprovalElection,
                         inner_distance: Callable) -> (float, list):
    """ Return: approvalwise distance """
    election_1.votes_to_approvalwise_vector()
    election_2.votes_to_approvalwise_vector()
    return inner_distance(election_1.approvalwise_vector, election_2.approvalwise_vector), None


def compute_hamming(election_1: ApprovalElection, election_2: ApprovalElection) -> float:
    """ Return: Hamming distance """
    votes_1 = election_1.votes
    votes_2 = election_2.votes
    params = {'voters': election_1.num_voters, 'candidates': election_2.num_candidates}
    file_name = f'{np.random.random()}.lp'
    path = os.path.join(os.getcwd(), "trash", file_name)
    lp.generate_ilp_distance(path, votes_1, votes_2, params, 'hamming')
    objective_value = lp.solve_ilp_distance(path, votes_1, votes_2, params, 'hamming')
    objective_value /= election_1.num_candidates  # ANALYZE THIS LINE
    lp.remove_lp_file(path)
    return objective_value


# # # # # # # # # # # # # # # #
# LAST CLEANUP ON: 13.10.2021 #
# # # # # # # # # # # # # # # #



import math

import numpy as np
import itertools


def kemeny_ranking(election):
    m = election.num_candidates
    wmg = election.votes_to_pairwise_matrix()
    best_d = np.infty
    for test_ranking in itertools.permutations(list(range(m))):
        dist = 0
        for i in range(m):
            for j in range(i + 1, m):
                dist = dist + wmg[test_ranking[j], test_ranking[i]]
            if dist > best_d:
                break
        if dist < best_d:
            best = test_ranking
            best_d = dist
    return best, best_d


def gini_coef(x):
    # Mean absolute difference
    mad = np.abs(np.subtract.outer(x, x)).mean()
    return (mad / x.mean() / 2)


def geom_mean(x):
    x = np.log(x)
    return np.exp(x.mean())


def swap_distance_between_potes(pote_1: list, pote_2: list) -> int:
    """ Return: Swap distance between two potes """
    swap_distance = 0
    for i, j in itertools.combinations(pote_1, 2):
        if (pote_1[i] > pote_1[j] and
            pote_2[i] < pote_2[j]) or \
                (pote_1[i] < pote_1[j] and
                 pote_2[i] > pote_2[j]):
            swap_distance += 1
    return swap_distance


def remove_diag(mtrx):
    """ Return: Input matrix with diagonal removed (shape[1] - 1) """
    res = np.zeros((mtrx.shape[0], mtrx.shape[1] - 1))
    for i in range(mtrx.shape[0]):
        for j in range(mtrx.shape[0]):
            if j < i:
                res[i, j] = mtrx[i, j]
            elif j > i:
                res[i, j - 1] = mtrx[i, j]
    return res


def calculate_borda_scores(election):
    m = election.num_candidates
    borda = np.zeros(m, int)
    for v in election.votes:
        for i, c in enumerate(v):
            borda[c] = borda[c] + m - i - 1
    return borda


def calculate_cand_dom_dist(election):
    distances = election.votes_to_pairwise_matrix()
    distances = np.abs(distances - 0.5)
    np.fill_diagonal(distances, 0)
    return distances


def calculate_cand_pos_dist(election):
    election.compute_potes()
    distances = np.zeros([election.num_candidates, election.num_candidates])
    for c1 in range(election.num_candidates):
        for c2 in range(election.num_candidates):
            dist = 0
            for pote in election.potes:
                dist += abs(pote[c1] - pote[c2])
            distances[c1][c2] = dist
    return distances


def calculate_vote_swap_dist(election):
    election.compute_potes()
    distances = np.zeros([election.num_voters, election.num_voters])
    for v1 in range(election.num_voters):
        for v2 in range(election.num_voters):
            distances[v1][v2] = swap_distance_between_potes(
                election.potes[v1], election.potes[v2])
    return distances


# DIVERSITY INDICES

def borda_gini(election) -> dict:
    if election.fake:
        return {'value': None}
    all_scores = calculate_borda_scores(election)
    return {'value': gini_coef(all_scores)}


def borda_meandev(election) -> dict:
    if election.fake:
        return {'value': None}
    all_scores = calculate_borda_scores(election)
    all_scores = np.abs(all_scores - all_scores.mean())
    return {'value': all_scores.mean()}


def borda_std(election) -> dict:
    if election.fake:
        return {'value': None}
    all_scores = calculate_borda_scores(election)
    return {'value': all_scores.std()}


def borda_range(election) -> dict:
    if election.fake:
        return {'value': None}
    all_scores = calculate_borda_scores(election)
    return (np.max(all_scores) - np.min(all_scores))


def cand_dom_dist_mean(election) -> dict:
    if election.fake:
        return {'value': None}
    distances = calculate_cand_dom_dist(election)
    return {'value': distances.sum() / (election.num_candidates - 1) / election.num_candidates * 2}


def agreement_index(election) -> dict:
    if election.fake:
        return {'value': None}
    distances = calculate_cand_dom_dist(election)
    return {'value': distances.sum() / (election.num_candidates - 1) / election.num_candidates * 2}


def cand_dom_dist_std(election) -> dict:
    if election.fake:
        return {'value': None}
    distances = calculate_cand_dom_dist(election)
    distances = remove_diag(distances)
    return {'value': distances.std()}


def cand_pos_dist_std(election) -> dict:
    if election.fake:
        return {'value': None}
    distances = calculate_cand_pos_dist(election)
    distances = remove_diag(distances)
    return {'value': distances.std()}


def cand_pos_dist_meandev(election) -> dict:
    if election.fake:
        return {'value': None}
    distances = calculate_cand_pos_dist(election)
    distances = remove_diag(distances)
    distances = np.abs(distances - distances.mean())
    return {'value': distances.mean()}


def cand_pos_dist_gini(election) -> dict:
    if election.fake:
        return {'value': None}
    distances = calculate_cand_pos_dist(election)
    distances = remove_diag(distances)
    return {'value': gini_coef(distances)}


def med_cands_summed(election) -> dict:
    if election.fake:
        return {'value': None}
    m = election.num_candidates
    distances = calculate_cand_pos_dist(election)
    res = [0] * m
    for i in range(1, m):
        best_d = np.inf
        for comb in itertools.combinations(range(m), i):
            d_total = 0
            for c1 in range(m):
                min_d = np.inf
                for c2 in comb:
                    d_cand = distances[c1, c2]
                    if d_cand < min_d:
                        min_d = d_cand
                d_total = d_total + min_d
            if d_total < best_d:
                best_d = d_total
        res[i] = best_d
    return {'value': sum(res)}


def vote_dist_mean(election) -> dict:
    if election.fake:
        return {'value': None}
    distances = calculate_vote_swap_dist(election)
    return {'value': distances.sum() / election.num_voters / (election.num_voters - 1)}


def vote_dist_max(election) -> dict:
    if election.fake:
        return {'value': None}
    distances = calculate_vote_swap_dist(election)
    return {'value': distances.max()}


def vote_dist_med(election) -> dict:
    if election.fake:
        return {'value': None}
    distances = calculate_vote_swap_dist(election)
    distances = remove_diag(distances)
    return {'value': np.median(distances)}


def vote_dist_gini(election) -> dict:
    if election.fake:
        return {'value': None}
    distances = calculate_vote_swap_dist(election)
    distances = remove_diag(distances)
    return {'value': gini_coef(distances)}


def vote_sqr_dist_mean(election) -> dict:
    if election.fake:
        return {'value': None}
    distances = calculate_vote_swap_dist(election)
    distances = remove_diag(distances)
    distances = np.sqrt(distances)
    return {'value': distances.mean()}


def vote_sqr_dist_med(election) -> dict:
    if election.fake:
        return {'value': None}
    distances = calculate_vote_swap_dist(election)
    distances = remove_diag(distances)
    distances = np.sqrt(distances)
    return {'value': np.median(distances)}


def vote_diversity_Karpov(election):
    if election.fake:
        return {'value': None}
    distances = calculate_vote_swap_dist(election)
    distances = remove_diag(distances)
    distances = distances + 0.5
    distances[distances == 0.5] = 1
    return {'value': geom_mean(distances)}


def dist_to_Kemeny_mean(election):
    if election.fake:
        return {'value': None}
    _, dist = kemeny_ranking(election)
    return dist / election.num_voters


def dist_to_Borda_mean(election) -> dict:
    if election.fake:
        return {'value': None}
    m = election.num_candidates
    borda = calculate_borda_scores(election)
    ranking = np.argsort(-borda)
    wmg = election.votes_to_pairwise_matrix()
    dist = 0
    for i in range(m):
        for j in range(i + 1, m):
            dist = dist + wmg[ranking[j], ranking[i]]
    return {'value': dist / election.num_voters}


def lexi_diversity(election):
    if election.fake:
        return {'value': None}
    return {'value': None}


def greedy_kKemenys_summed(election) -> dict:
    if election.fake:
        return {'value': None}
    res = [0] * election.num_voters
    distances = calculate_vote_swap_dist(election)
    best = np.argmin(distances.sum(axis=1))
    best_vec = distances[best]
    res[0] = best_vec.sum()
    distances = np.vstack((distances[:best], distances[best + 1:]))

    for i in range(1, election.num_voters):
        relatives = distances - best_vec
        relatives = relatives * (relatives < 0)
        best = np.argmin(relatives.sum(axis=1))
        best_vec = best_vec + relatives[best]
        res[i] = best_vec.sum()
        distances = np.vstack((distances[:best], distances[best + 1:]))

    return sum(res)
    max_dist = (election.num_candidates) * (election.num_candidates - 1) / 2
    return {'value': sum(res) / election.num_voters / max_dist}


def restore_order(x):
    for i in range(len(x)):
        for j in range(len(x) - i, len(x)):
            if x[j] >= x[-i - 1]:
                x[j] += 1
    return x


def distances_to_rankings(rankings, distances):
    dists = distances[rankings]
    return np.sum(dists.min(axis=0))


def find_improvement(distances, d, starting, rest, n, k, l):
    for cut in itertools.combinations(range(k), l):
        # print(cut)
        for paste in itertools.combinations(rest, l):
            ranks = []
            j = 0
            for i in range(k):
                if i in cut:
                    ranks.append(paste[j])
                    j = j + 1
                else:
                    ranks.append(starting[i])
            # check if unique
            if len(set(ranks)) == len(ranks):
                # check if better
                d_new = distances_to_rankings(ranks, distances)
                if d > d_new:
                    return ranks, d_new, True
    return starting, d, False


def local_search_kKemeny_single_k(election, k, l, starting=None) -> dict:
    if starting is None:
        starting = list(range(k))
    distances = calculate_vote_swap_dist(election)

    n = election.num_voters

    d = distances_to_rankings(starting, distances)
    iter = 0
    check = True
    while (check):
        # print(iter)
        # print(starting)
        # print(d)
        # print()
        iter = iter + 1
        rest = [i for i in range(n) if i not in starting]
        for j in range(l):
            starting, d, check = find_improvement(distances, d, starting, rest, n, k, j + 1)
            if check:
                break
        # print()
    return {'value': d}


def local_search_kKemeny(election, l, starting=None) -> dict:
    max_dist = election.num_candidates * (election.num_candidates - 1) / 2
    res = []
    for k in range(1, election.num_voters):
        # print(k)
        if starting is None:
            d = local_search_kKemeny_single_k(election, k, l)['value']
        else:
            d = local_search_kKemeny_single_k(election, k, l, starting[:k])['value']
        d = d / max_dist / election.num_voters
        if d > 0:
            res.append(d)
        else:
            break
    for k in range(len(res), election.num_voters):
        res.append(0)

    return {'value': res}


def diversity_index(election) -> dict:
    if election.fake:
        return {'value': None}
    max_dist = election.num_candidates * (election.num_candidates - 1) / 2
    res = [0] * election.num_voters
    chosen_votes = []
    distances = calculate_vote_swap_dist(election)
    best = np.argmin(distances.sum(axis=1))
    chosen_votes.append(best)
    best_vec = distances[best]
    res[0] = best_vec.sum() / max_dist / election.num_voters
    distances = np.vstack((distances[:best], distances[best + 1:]))

    for i in range(1, election.num_voters):
        relatives = distances - best_vec
        relatives = relatives * (relatives < 0)
        best = np.argmin(relatives.sum(axis=1))
        chosen_votes.append(best)
        best_vec = best_vec + relatives[best]
        res[i] = best_vec.sum() / max_dist / election.num_voters
        distances = np.vstack((distances[:best], distances[best + 1:]))

    chosen_votes = restore_order(chosen_votes)

    res_1 = local_search_kKemeny(election, 1, chosen_votes)['value']
    res_2 = local_search_kKemeny(election, 1)['value']
    res = [min(d_1, d_2) for d_1, d_2 in zip(res_1, res_2)]

    return {'value': sum([x / (i + 1) for i, x in enumerate(res)])}


def greedy_kKemenys_divk_summed(election) -> dict:
    if election.fake:
        return {'value': None}
    res = [0] * election.num_voters
    distances = calculate_vote_swap_dist(election)
    best = np.argmin(distances.sum(axis=1))
    best_vec = distances[best]
    res[0] = best_vec.sum()
    distances = np.vstack((distances[:best], distances[best + 1:]))

    for i in range(1, election.num_voters):
        relatives = distances - best_vec
        relatives = relatives * (relatives < 0)
        best = np.argmin(relatives.sum(axis=1))
        best_vec = best_vec + relatives[best]
        res[i] = best_vec.sum() / (i + 1)
        distances = np.vstack((distances[:best], distances[best + 1:]))

    # res[0] = 0 # for disregarding one Kemeny (AN = ID)
    max_dist = (election.num_candidates) * (election.num_candidates - 1) / 2
    return {'value': sum(res) / election.num_voters / max_dist}


def greedy_2kKemenys_summed(election) -> dict:
    if election.fake:
        return {'value': None}
    res = []
    distances = calculate_vote_swap_dist(election)
    best = np.argmin(distances.sum(axis=1))
    best_vec = distances[best]
    res.append(best_vec.sum())
    distances = np.vstack((distances[:best], distances[best + 1:]))

    k = 2
    for i in range(1, election.num_voters):
        relatives = distances - best_vec
        relatives = relatives * (relatives < 0)
        best = np.argmin(relatives.sum(axis=1))
        best_vec = best_vec + relatives[best]
        if (i + 1) == k:
            res.append(best_vec.sum())
            k = k * 2
        distances = np.vstack((distances[:best], distances[best + 1:]))

    # res[0] = 0 # for disregarding one Kemeny (AN = ID)
    max_dist = (election.num_candidates) * (election.num_candidates - 1) / 2
    return {'value': sum(res) / election.num_voters / max_dist / 2}


def polarization_1by2Kemenys(election) -> dict:
    if election.fake:
        return {'value': None}
    distances = calculate_vote_swap_dist(election)
    best = np.argmin(distances.sum(axis=1))
    best_vec = distances[best]
    first_kemeny = best_vec.sum()
    distances = np.vstack((distances[:best], distances[best + 1:]))

    relatives = distances - best_vec
    relatives = relatives * (relatives < 0)
    best = np.argmin(relatives.sum(axis=1))
    best_vec = best_vec + relatives[best]
    second_kemeny = best_vec.sum()

    max_dist = (election.num_candidates) * (election.num_candidates - 1) / 2
    return {'value': (first_kemeny - second_kemeny) / election.num_voters / max_dist}


def polarization_index(election) -> dict:
    if election.fake:
        return {'value': None}
    distances = calculate_vote_swap_dist(election)
    best_1 = np.argmin(distances.sum(axis=1))
    best_vec = distances[best_1]
    first_kemeny = best_vec.sum()
    distances = np.vstack((distances[:best_1], distances[best_1 + 1:]))

    relatives = distances - best_vec
    relatives = relatives * (relatives < 0)
    best_2 = np.argmin(relatives.sum(axis=1))

    if best_1 <= best_2:
        best_2 = best_2 + 1

    chosen = [best_1, best_2]
    chosen.sort()

    second_kemeny_1 = local_search_kKemeny_single_k(election, 2, 1, starting=chosen)['value']
    second_kemeny_2 = local_search_kKemeny_single_k(election, 2, 1)['value']
    second_kemeny = min(second_kemeny_1, second_kemeny_2)

    max_dist = (election.num_candidates) * (election.num_candidates - 1) / 2
    return {'value': 2 * (first_kemeny - second_kemeny) / election.num_voters / max_dist}


def greedy_kmeans_summed(election) -> dict:
    if election.fake:
        return {'value': None}
    res = [0] * election.num_voters
    distances = calculate_vote_swap_dist(election)
    distances = distances * distances
    best = np.argmin(distances.sum(axis=1))
    best_vec = distances[best]
    res[0] = best_vec.sum()
    distances = np.vstack((distances[:best], distances[best + 1:]))

    for i in range(1, election.num_voters):
        relatives = distances - best_vec
        relatives = relatives * (relatives < 0)
        best = np.argmin(relatives.sum(axis=1))
        best_vec = best_vec + relatives[best]
        res[i] = best_vec.sum()
        distances = np.vstack((distances[:best], distances[best + 1:]))

    return {'value': sum(res)}


def support_diversity(election, tuple_len) -> dict:
    if election.fake:
        return {'value': None}
    m = election.num_candidates
    res = 0
    for subset in itertools.combinations(range(m), tuple_len):
        support = []
        for v in election.votes:
            trimmed_v = []
            for c in v:
                if c in subset:
                    trimmed_v.append(c)
            if not (trimmed_v in support):
                support.append(trimmed_v)
        res = res + len(support)
    return {'value': res}


def support_diversity_normed(election, tuple_len) -> dict:
    if election.fake:
        return {'value': None}
    m = election.num_candidates
    res = 0
    count = 0
    for subset in itertools.combinations(range(m), tuple_len):
        count = count + 1
        support = []
        for v in election.votes:
            trimmed_v = []
            for c in v:
                if c in subset:
                    trimmed_v.append(c)
            if not (trimmed_v in support):
                support.append(trimmed_v)
        res = res + len(support)
    return {'value': res / count}


def support_diversity_normed2(election, tuple_len) -> dict:
    if election.fake:
        return {'value': None}
    m = election.num_candidates
    res = 0
    count = 0
    for subset in itertools.combinations(range(m), tuple_len):
        count = count + 1
        support = []
        for v in election.votes:
            trimmed_v = []
            for c in v:
                if c in subset:
                    trimmed_v.append(c)
            if not (trimmed_v in support):
                support.append(trimmed_v)
        res = res + len(support)
    return {'value': res / count / math.factorial(tuple_len)}


def support_diversity_normed3(election, tuple_len) -> dict:
    if election.fake:
        return {'value': None}
    m = election.num_candidates
    res = 0
    count = 0
    for subset in itertools.combinations(range(m), tuple_len):
        count = count + 1
        support = []
        for v in election.votes:
            trimmed_v = []
            for c in v:
                if c in subset:
                    trimmed_v.append(c)
            if not (trimmed_v in support):
                support.append(trimmed_v)
        res = res + len(support)
    max_times = min(math.factorial(tuple_len), election.num_voters)
    return {'value': res / count / max_times}


def support_pairs(election):
    return support_diversity(election, 2)


def support_triplets(election):
    return support_diversity(election, 3)


def support_votes(election) -> dict:
    if election.fake:
        return {'value': None}
    m = election.num_candidates
    return {'value': support_diversity(election, m)}


def support_diversity_summed(election) -> dict:
    if election.fake:
        return {'value': None}
    m = election.num_candidates
    res = 0
    for i in range(2, m + 1):
        res = res + support_diversity(election, i)['value']
    return {'value': res}


def support_diversity_normed_summed(election) -> dict:
    if election.fake:
        return {'value': None}
    m = election.num_candidates
    res = 0
    for i in range(2, m + 1):
        res = res + support_diversity_normed(election, i)['value']
    return {'value': res}


def support_diversity_normed2_summed(election) -> dict:
    if election.fake:
        return {'value': None}
    m = election.num_candidates
    res = 0
    for i in range(2, m + 1):
        res = res + support_diversity_normed2(election, i)['value']
    return {'value': res}


def support_diversity_normed3_summed(election) -> dict:
    if election.fake:
        return {'value': None}
    m = election.num_candidates
    res = 0
    for i in range(2, m + 1):
        res = res + support_diversity_normed3(election, i)['value']
    return {'value': res}



import prefsampling.ordinal as pref_ordinal


def gs_mask(num_voters=None,
            num_candidates=None,
            tree_sampler=None,
            **kwargs):

    if type(tree_sampler) is str:
        if tree_sampler.lower() == 'balanced':
            tree_sampler = pref_ordinal.TreeSampler.BALANCED
        elif tree_sampler.lower() == 'caterpillar':
            tree_sampler = pref_ordinal.TreeSampler.CATERPILLAR

    if tree_sampler is None:
        tree_sampler = pref_ordinal.TreeSampler.SCHROEDER

    return pref_ordinal.group_separable(num_voters=num_voters,
                                        num_candidates=num_candidates,
                                        tree_sampler=tree_sampler,
                                        **kwargs)



import copy
import os
import pickle
import random

import numpy as np
import logging

import mapel.core.features.mallows as ml


def generate_mallows_votes(*args, **kwargs):
    return ml.generate_mallows_votes(*args, **kwargs)


def generate_norm_mallows_mixture_votes(num_voters, num_candidates, params):
    phi_1 = ml.phi_from_normphi(num_candidates, float(params['normphi_1']))
    params_1 = {'weight': 0, 'phi': phi_1}
    votes_1 = generate_mallows_votes(num_voters, num_candidates, params_1)

    phi_2 = ml.phi_from_normphi(num_candidates, float(params['normphi_2']))
    params_2 = {'weight': 1, 'phi': phi_2}
    votes_2 = generate_mallows_votes(num_voters, num_candidates, params_2)

    votes = []
    size_1 = int((1 - float(params['weight'])) * num_voters)
    for i in range(size_1):
        votes.append(votes_1[i])
    for i in range(size_1, num_voters):
        votes.append(votes_2[i])

    return votes


def calculateZpoly(m):
    res = [1]
    for i in range(1, m + 1):
        mult = [1] * i
        res2 = [0] * (len(res) + len(mult) - 1)
        for o1, i1 in enumerate(res):
            for o2, i2 in enumerate(mult):
                res2[o1 + o2] += i1 * i2
        res = res2
    return res


def evaluatePolynomial(coeff, x):
    res = 0
    for i, c in enumerate(coeff):
        res += c * (x ** i)
    return res


def calculateZ(m, phi):
    coeff = calculateZpoly(m)
    return evaluatePolynomial(coeff, phi)


# mat[i][j] is the probability with which candidate i ends up in position j
def mallowsMatrix(num_candidates, lphi, pos, normalize=True):
    mat = np.zeros([num_candidates, num_candidates])
    if normalize:
        phi = ml.phi_from_normphi(num_candidates, lphi)
    else:
        phi = lphi
    Z = calculateZ(num_candidates, phi)
    for i in range(num_candidates):
        for j in range(num_candidates):
            freqs = [pos[k][i][j] for k in
                     range(1 + int(num_candidates * (num_candidates - 1) / 2))]
            unnormal_prob = evaluatePolynomial(freqs, phi)
            mat[i][j] = unnormal_prob / Z
    return mat


def get_mallows_matrix(num_candidates, params, normalize=True):
    lphi = params['normphi']
    if 'weight' not in params:
        weight = 0
    else:
        weight = params['weight']

    if 'sec_normphi' not in params:
        lphi_2 = lphi
    else:
        lphi_2 = params['sec_normphi']

    try:
        path = os.path.join(os.getcwd(), 'mapel', 'elections', 'cultures',
                            'mallows_positionmatrices',
                            str(num_candidates) + "_matrix.txt")
        print(path)
        with open(path, "rb") as file:
            pos = pickle.load(file)
    except FileNotFoundError:
        print("Mallows matrix only supported for up to 30 candidates")
    mat1 = mallowsMatrix(num_candidates, lphi, pos, normalize)
    mat2 = mallowsMatrix(num_candidates, lphi_2, pos, normalize)
    res = np.zeros([num_candidates, num_candidates])
    for i in range(num_candidates):
        for j in range(num_candidates):
            res[i][j] = (1. - weight) * mat1[i][j] + (weight) * mat2[i][num_candidates - 1 - j]
    return res


def get_mallows_vectors(num_candidates, fake_param):
    return get_mallows_matrix(num_candidates, fake_param).transpose()


def generate_mallows_party(num_voters=None,
                           num_candidates=None,
                           election_model=None,
                           params=None):
    num_parties = params['num_parties']
    num_winners = params['num_winners']
    party_size = num_winners

    params['phi'] = ml.phi_from_normphi(num_parties, normphi=params['main-phi'])
    mapping = generate_mallows_votes(num_voters, num_parties, params)[0]

    params['phi'] = ml.phi_from_normphi(num_parties, normphi=params['normphi'])
    votes = generate_mallows_votes(num_voters, num_parties, params)

    for i in range(num_voters):
        for j in range(num_parties):
            votes[i][j] = mapping[votes[i][j]]

    new_votes = [[] for _ in range(num_voters)]

    for i in range(num_voters):
        for j in range(num_parties):
            for w in range(party_size):
                _id = votes[i][j] * party_size + w
                new_votes[i].append(_id)

    return new_votes


def generate_approval_truncated_mallows_votes(num_voters=None,
                                              num_candidates=None,
                                              max_range=1,
                                              normphi=None,
                                              weight=None,
                                              **kwargs):
    phi = ml.phi_from_normphi(num_candidates, normphi=normphi)

    ordinal_votes = generate_mallows_votes(num_voters, num_candidates, phi=phi, weight=weight)

    votes = []
    k = np.random.randint(low=1., high=int(max_range * num_candidates))
    for v in range(num_voters):
        votes.append(set(ordinal_votes[v][0:k]))

    return votes


def runif_in_simplex(n):
    ''' Return uniformly random vector in the n-simplex '''

    k = np.random.exponential(scale=1.0, size=n)
    return k / sum(k)




import sys
import os

try:
    from dotenv import load_dotenv
    load_dotenv()
    # sys.path.append(os.environ["PATH"])
    from abcvoting.preferences import Profile
    from abcvoting import abcrules, properties
    from abcvoting.output import output, INFO
except ImportError:
    pass


def test_ejr(election, rule):

    print(election.election_id, rule)

    profile = Profile(election.num_candidates)
    profile.add_voters(election.votes)
    committee = election.winning_committee[rule]

    return properties.full_analysis(profile, committee)




from collections import Counter

from .distances_ import get_distance
import mapel.core.printing as pr
from .objects.ApprovalElectionExperiment import ApprovalElectionExperiment
from .objects.OrdinalElection import OrdinalElection
from .objects.ApprovalElection import ApprovalElection
from .objects.OrdinalElectionExperiment import OrdinalElectionExperiment


def prepare_online_ordinal_experiment(**kwargs):
    return prepare_experiment(**kwargs, instance_type='ordinal',
                              is_exported=False, is_imported=False)


def prepare_offline_ordinal_experiment(**kwargs):
    return prepare_experiment(**kwargs, instance_type='ordinal',
                              is_exported=True, is_imported=True)


def prepare_online_approval_experiment(**kwargs):
    return prepare_experiment(**kwargs, instance_type='approval',
                              is_exported=False, is_imported=False)


def prepare_offline_approval_experiment(**kwargs):
    return prepare_experiment(**kwargs, instance_type='approval',
                              is_exported=True, is_imported=True)


def prepare_experiment(experiment_id=None,
                       instances=None,
                       distances=None,
                       instance_type=None,
                       coordinates=None,
                       distance_id=None,
                       is_imported=False,
                       is_shifted=False,
                       is_exported=True,
                       coordinates_names=None,
                       embedding_id=None,
                       fast_import=False,
                       with_matrix=False,
                       dim=2):
    if instance_type == 'ordinal':
        return OrdinalElectionExperiment(experiment_id=experiment_id,
                                         is_shifted=is_shifted,
                                         instances=instances,
                                         is_exported=is_exported,
                                         is_imported=is_imported,
                                         distances=distances,
                                         coordinates=coordinates,
                                         distance_id=distance_id,
                                         coordinates_names=coordinates_names,
                                         embedding_id=embedding_id,
                                         fast_import=fast_import,
                                         with_matrix=with_matrix,
                                         instance_type=instance_type,
                                         dim=dim)
    elif instance_type in ['approval', 'rule']:
        return ApprovalElectionExperiment(experiment_id=experiment_id,
                                          is_shifted=is_shifted,
                                          instances=instances,
                                          is_exported=is_exported,
                                          is_imported=is_imported,
                                          distances=distances,
                                          coordinates=coordinates,
                                          distance_id=distance_id,
                                          coordinates_names=coordinates_names,
                                          embedding_id=embedding_id,
                                          fast_import=fast_import,
                                          instance_type=instance_type,
                                          dim=dim)


def print_approvals_histogram(*args):
    pr.print_approvals_histogram(*args)


def custom_div_cmap(**kwargs):
    return pr.custom_div_cmap(**kwargs)


def print_matrix(**kwargs):
    pr.print_matrix(**kwargs)


### WITHOUT EXPERIMENT ###
def generate_election(**kwargs):
    election = OrdinalElection(**kwargs)
    election.prepare_instance()
    return election


def generate_ordinal_election(**kwargs):
    election = OrdinalElection(**kwargs, is_exported=False)
    election.prepare_instance()
    return election


def generate_approval_election(**kwargs):
    election = ApprovalElection(**kwargs, is_exported=False)
    election.prepare_instance()
    return election


def generate_election_from_votes(votes=None):
    election = OrdinalElection()
    election.num_candidates = len(votes[0])
    election.num_voters = len(votes)
    election.votes = votes
    return election


def generate_ordinal_election_from_votes(votes=None):
    election = OrdinalElection()
    election.num_candidates = len(votes[0])
    election.num_voters = len(votes)
    election.votes = votes
    election.is_exported = False
    c = Counter(map(tuple, votes))
    counted_votes = [[count, list(row)] for row, count in c.items()]
    counted_votes = sorted(counted_votes, reverse=True)
    election.quantites = [a[0] for a in counted_votes]
    election.distinct_votes = [a[1] for a in counted_votes]
    election.num_options = len(counted_votes)
    return election


def generate_approval_election_from_votes(votes=None, num_candidates=None):
    election = ApprovalElection()
    if num_candidates is None:
        election.num_candidates = len(set().union(*votes))
    else:
        election.num_candidates = num_candidates
    election.num_voters = len(votes)
    election.votes = votes
    election.is_exported = False
    c = Counter(map(tuple, votes))
    counted_votes = [[count, list(row)] for row, count in c.items()]
    counted_votes = sorted(counted_votes, reverse=True)
    election.quantites = [a[0] for a in counted_votes]
    election.distinct_votes = [a[1] for a in counted_votes]
    election.num_options = len(counted_votes)
    return election


def compute_distance(*args, **kwargs):
    return get_distance(*args, **kwargs)



#!/usr/bin/env python
import logging
from abc import ABC
from collections import Counter

from matplotlib import pyplot as plt
from mapel.elections.cultures_ import generate_approval_votes
from mapel.elections.objects.Election import Election
from mapel.core.inner_distances import hamming
from mapel.core.utils import *
import mapel.elections.persistence.election_imports as imports
import mapel.elections.persistence.election_exports as exports
from mapel.elections.cultures.params import *


class ApprovalElection(Election, ABC):

    def __init__(self,
                 experiment_id=None,
                 election_id=None,
                 culture_id=None,
                 num_candidates=None,
                 variable=None,
                 **kwargs):

        super().__init__(experiment_id=experiment_id,
                         election_id=election_id,
                         culture_id=culture_id,
                         num_candidates=num_candidates,
                         **kwargs)

        self.variable = variable

        self.approvalwise_vector = []
        self.reverse_approvals = []

        self.import_approval_election()

    def import_approval_election(self):
        if self.is_imported and self.experiment_id is not None:
            try:
                fake = imports.check_if_fake(self.experiment_id, self.election_id, 'app')
                if fake:
                    self.culture_id, self.params, self.num_voters, self.num_candidates = \
                        imports.import_fake_app_election(self.experiment_id, self.election_id)
                else:
                    self.votes, self.num_voters, self.num_candidates, self.params, \
                        self.culture_id, self.num_options, self.quantites, self.distinct_votes \
                        = imports.import_real_app_election(
                            experiment_id=self.experiment_id,
                            election_id=self.election_id,
                            is_shifted=self.is_shifted)
            except:
                pass

        if self.params is None:
            self.params = {}

        if self.culture_id is not None:
            self.params, self.printing_params = update_params_approval(self.params,
                                                                       self.printing_params,
                                                                       self.variable,
                                                                       self.culture_id,
                                                                       self.num_candidates)

    def votes_to_approvalwise_vector(self) -> None:
        """ Convert votes to approvalwise vectors """
        approvalwise_vector = np.zeros([self.num_candidates])
        for vote in self.votes:
            for c in vote:
                approvalwise_vector[c] += 1
        approvalwise_vector = approvalwise_vector / self.num_voters
        self.approvalwise_vector = np.sort(approvalwise_vector)

    def compute_reverse_approvals(self):
        self.reverse_approvals = [set(i for i, vote in enumerate(self.votes) if c in vote)
                                  for c in range(self.num_candidates)]

    def prepare_instance(self, is_exported=None, is_aggregated=True):
        self.votes = generate_approval_votes(culture_id=self.culture_id,
                                             num_candidates=self.num_candidates,
                                             num_voters=self.num_voters,
                                             params=self.params)
        if not self.fake:
            c = Counter(map(tuple, self.votes))
            counted_votes = [[count, list(row)] for row, count in c.items()]
            counted_votes = sorted(counted_votes, reverse=True)
            self.quantites = [a[0] for a in counted_votes]
            self.distinct_votes = [a[1] for a in counted_votes]
            self.num_options = len(counted_votes)
        else:
            self.quantites = [self.num_voters]
            self.num_options = 1

        if is_exported:
            exports.export_approval_election(self, is_aggregated=is_aggregated)

    def _compute_distances_between_votes(self, distance_id='hamming'):
        distances = np.zeros([self.num_voters, self.num_voters])
        for v1 in range(self.num_voters):
            for v2 in range(self.num_voters):
                if distance_id == 'hamming':
                    distances[v1][v2] = hamming(self.votes[v1], self.votes[v2])
                elif distance_id == 'jaccard':
                    if len(self.votes[v1].union(self.votes[v2])) == 0:
                        distances[v1][v2] = 1
                    else:
                        distances[v1][v2] = 1 - len(
                            self.votes[v1].intersection(self.votes[v2])) / len(
                            self.votes[v1].union(self.votes[v2]))

        self.distances['vote'] = distances

        if self.is_exported:
            exports.export_distances(self, object_type='vote')

        return distances

    def _compute_distances_between_candidates(self, distance_id='hamming'):

        distances = np.zeros([self.num_candidates, self.num_candidates])
        for c1 in range(self.num_candidates):
            for c2 in range(self.num_candidates):
                if distance_id == 'hamming':
                    distances[c1][c2] = hamming(self.reverse_approvals[c1],
                                                self.reverse_approvals[c2])
                elif distance_id == 'jaccard':
                    if len(self.reverse_approvals[c1].union(self.reverse_approvals[c2])) == 0:
                        distances[c1][c2] = 1
                    else:
                        distances[c1][c2] = 1 - len(
                            self.reverse_approvals[c1].intersection(self.reverse_approvals[c2])) \
                                            / len(
                            self.reverse_approvals[c1].union(self.reverse_approvals[c2]))

        self.distances['candidate'] = distances

        if self.is_exported:
            exports.export_distances(self, object_type='candidate')

        return distances

    def compute_distances(self, object_type=None, distance_id='hamming'):
        if object_type == 'vote':
            return self._compute_distances_between_votes(distance_id=distance_id)
        elif object_type == 'candidate':
            return self._compute_distances_between_candidates(distance_id=distance_id)

    def print_map(
            self,
            show=True,
            radius=None,
            name=None,
            alpha=0.1,
            s=30,
            object_type=None,
            double_gradient=False,
            saveas=None,
            color='blue',
            marker='o',
            title_size=20,
            annotate=False
    ):

        if object_type == 'vote':
            length = self.num_voters
            votes = self.votes
        elif object_type == 'candidate':
            length = self.num_candidates
            self.compute_reverse_approvals()
            votes = self.reverse_approvals
        else:
            logging.warning(f'Incorrect object type: {object_type}')
            length = 0

        if object_type is None:
            object_type = self.object_type

        plt.figure(figsize=(6.4, 6.4))

        X = []
        Y = []
        for elem in self.coordinates[object_type]:
            X.append(elem[0])
            Y.append(elem[1])

        start = False
        if start:
            plt.scatter(X[0], Y[0],
                        color='sienna',
                        s=1000,
                        alpha=1,
                        marker='X')

        if double_gradient:
            for i in range(length):
                x = float(self.points['voters'][i][0])
                y = float(self.points['voters'][i][1])
                plt.scatter(X[i], Y[i], color=[0, y, x], s=s, alpha=alpha)
        else:
            plt.scatter(X, Y, color=color, s=s, alpha=alpha, marker=marker)

        if annotate:
            for i in range(len(X)):
                plt.annotate(i, (X[i], Y[i]), color='black')

        avg_x = np.mean(X)
        avg_y = np.mean(Y)

        if radius:
            plt.xlim([avg_x - radius, avg_x + radius])
            plt.ylim([avg_y - radius, avg_y + radius])

        plt.axis('off')

        if saveas is not None:

            if saveas == 'default':

                saveas = f'{self.label}_{object_type}'
                file_name = os.path.join(os.getcwd(), "images", name, f'{saveas}.png')
                plt.savefig(file_name, bbox_inches='tight', dpi=100)

        if show:
            plt.show()
        else:
            plt.clf()

        plt.close()



#!/usr/bin/env python

from numpy import ceil
import time

try:
    import pulp
except Exception:
    pulp = None


# from mapel.elections.objects.ApprovalElection import ApprovalElection
from math import ceil
import itertools
from collections import defaultdict


def count_number_of_cohesive_groups_brute(election, l: int = 1,
                                          committee_size: int = 10):
    answer = 0
    min_size = int(ceil(l * election.num_voters / committee_size))
    voters = [i for i in range(0, election.num_voters)]
    for s in powerset(voters, min_size=min_size):
        if len(s) < min_size:
            continue
        cands = set(election.votes[s[0]])
        for v in s:
            cands &= election.votes[v]
        if len(cands) >= l:
            answer += 1
    return answer


####################################################################################################
####################################################################################################
####################################################################################################

def powerset(iterable, min_size=0):
    "powerset([1,2,3]) --> () (1,) (2,) (3,) (1,2) (1,3) (2,3) (1,2,3)"
    s = list(iterable)
    return itertools.chain.from_iterable(
        itertools.combinations(s, r) for r in range(min_size, len(s) + 1))


def newton(n: int, k: int):
    if k > n:
        return 0
    answer = 1
    for i in range(n - k + 1, n + 1):
        answer *= i
    for i in range(1, k + 1):
        answer //= i
    return answer


def count_number_of_cohesive_groups(election, l: int = 1,
                                    committee_size: int = 10):

    if l > 1:
        raise NotImplementedError()
    answer = 0
    d = defaultdict(lambda: 0)
    timeout = time.time() + 20 * 1  # 20s from now
    for v in range(election.num_voters):
        for s in powerset(election.votes[v], min_size=1):
            d[s] += 1
            if time.time() > timeout:
                return -1
    min_size = int(ceil(l * election.num_voters / committee_size))
    for s in d:
        for siz in range(min_size, d[s] + 1):
            sign = 2 * (len(s) % 2) - 1  # 1 for even, -1 for odd, comes from (-1) ^ (s-1)
            answer += newton(d[s], siz) * sign
    return answer


####################################################################################################
####################################################################################################
####################################################################################################

def count_largest_cohesiveness_level_l_of_cohesive_group(election, feature_params):
    committee_size = feature_params['committee_size']

    # if election.model == 'approval_zeros':
    #     return 0
    # elif election.model == 'approval_ones':
    #     return min(committee_size, election.num_candidates)

    l_ans = 0
    for l in range(1, election.num_voters + 1):
        if solve_ilp_instance(election, committee_size, l):
            l_ans = l
        else:
            break
    return l_ans


def solve_ilp_instance(election, committee_size: int, l: int = 1) -> bool:
    pulp.getSolver('CPLEX_CMD')
    model = pulp.LpProblem("cohesiveness_level_l", pulp.LpMaximize)
    X = [pulp.LpVariable("x_" + str(i), cat='Binary') for i in
         range(election.num_voters)]  # X[i] = 1 if we select i-th voter, otherwise 0
    Y = [pulp.LpVariable("y_" + str(j), cat='Binary') for j in
         range(election.num_candidates)]  # Y[j] = 1 if we select j-th candidate, otherwise 0
    s = int(ceil(
        l * election.num_voters / committee_size))  # If there is any valid l-cohesive group, then there is also at least one with minimum possible size

    objective = l
    model += objective  # We want to maximize cohesiveness level l (but l is constant, only convention)

    x_sum_eq = 0
    for x in X:
        x_sum_eq += x
    model += x_sum_eq == s  # We choose exactly s voters

    y_sum_ineq = 0
    for y in Y:
        y_sum_ineq += y
    model += y_sum_ineq >= l  # We choose at least l candidates (although l are sufficient in this case)

    cand_to_voters_variables_list = [[] for j in range(election.num_candidates)]
    for i, d in enumerate(election.votes):
        for j in d:
            cand_to_voters_variables_list[j].append(X[i])
    # We want to assert that the selected voters approve all the selected candidates.
    # For each candidate j,  we construct the following inequality:  a_{0,j} * x_0 + a_{1,j} * x_1 + ... + a_{n-1,j} * x_{n-1}  -   s * y_j    >=    0
    # We define a_{i, j} as the flag indicating whether i-th voter approves j-th candidate (1 if yes, otherwise 0)
    # Let us observe that if the j-th candidate is not selected, then s * y_j = 0 and the above inequality is naturally satisfied.
    # However, if j-th candidate is selected, then the above can be satisfied if and only if all s selected voters approve j-th candidate
    for j, y in enumerate(Y):
        y_ineq = 0
        for x in cand_to_voters_variables_list[j]:
            y_ineq += x
        y_ineq -= s * y
        model += y_ineq >= 0

    model.solve(pulp.PULP_CBC_CMD(msg=False))
    # if LpStatus[culture_id.status] == 'Optimal':
    #     print([var.election_id + "=" + str(var.varValue) for var in culture_id.variables() if var.varValue is not None and var.varValue > 0], sep=" ")    # prints result variables which have value > 0
    return pulp.LpStatus[model.status] == 'Optimal'



from PIL import Image, ImageDraw

import mapel


def getrgb(value, MAX):
    x = int(255 * value / MAX)
    return (x, x, x)


def getrgb_uniform(value, MAX):
    x = int(255 * value)
    return (x, x, x)


def getsqrtrgb(value, MAX):
    x = int(255 * (value ** 0.33) / (MAX ** 0.33))
    return (x, x, x)


def getsqrtrgb_uniform(value, MAX):
    x = int(255 * (value ** 0.25))
    return (x, x, x)


def matrix2png(argv):
    # introduce yourself
    if len(argv) < 4:
        print("Invocation:")
        print("  python3 matrix2png num_candidates culture_id reorder [param1]")
        print(
            "  reorder -- election_id of the culture_id to try to resemble (e.g., ID, or AN); use org to use original order")
        print("")
        exit()

    # gather arguments
    m = int(argv[1])
    n = m * m
    model = argv[2]
    tgt = argv[3]
    print("TGT:", tgt)
    if len(argv) >= 5:
        param = float(argv[4])
    else:
        param = None

    if model != "mallows":
        name = "%s_%d_%s.png" % (model, m, tgt)
    else:
        name = "%s_phi%d_%d_%s.png" % (model, param * 100, m, tgt)

    # prepare the election/matrix
    experiment = mapel.prepare_experiment()
    experiment.set_default_num_candidates(m)
    experiment.set_default_num_voters(n)

    # Compass Matrices
    experiment.add_election(election_model="uniformity", election_id="UN", color=(1, 0.5, 0.5),
                            marker="X")
    experiment.add_election(election_model="identity", election_id="ID", color="red", marker="X")
    experiment.add_election(election_model="antagonism", election_id="AN", color="black",
                            marker="o")
    experiment.add_election(election_model="stratification", election_id="ST", color="black")

    # form the matrix
    if model != "mallows":
        experiment.add_election(election_model=model, election_id="M")
    else:
        experiment.add_election(election_model="norm-mallows_matrix", params={"normphi": param},
                                election_id="M")
    M = experiment.elections["M"].matrix

    # get the mapping to a given election
    experiment.compute_distances()
    if tgt == "org":
        match = list(range(m))
    else:
        match = experiment.matchings[tgt]["M"]
    print(match)
    # get reversed matching
    rev_match = [0] * m
    for i in range(m):
        rev_match[match[i]] = i
    print(rev_match)

    # create the image
    img = Image.new("RGB", (m, m), color="black")
    draw = ImageDraw.Draw(img)

    MAX = 0  # highest value in the matrix
    for y in range(m):
        for x in range(m):
            MAX = max(MAX, M[y][x])

    color = lambda v: getsqrtrgb_uniform(v, MAX)

    ### print columns
    print("----")
    for x in range(m):
        print("%.2f" % x, end=" ")
    print()
    print("----")

    ### draw the matrix
    for y in range(m):
        for x in range(m):
            draw.point((x, y), fill=color(M[y][rev_match[x]]))
            print("%.2f" % M[y][rev_match[x]], end=" ")
        print()

    # save the image
    img.save(name)

    print("MAX value:", MAX)




import numpy as np


def clustering_v1(experiment, num_clusters=20):
    from scipy.cluster.hierarchy import dendrogram, linkage, fcluster
    import scipy.spatial.distance as ssd

    # skip the paths

    SKIP = ['UNID', 'ANID', 'STID', 'ANUN', 'STUN', 'STAN',
            'Mallows',
            'Urn',
            'Identity', 'Uniformity', 'Antagonism', 'Stratification',
            ]

    new_names = []
    for i, a in enumerate(list(experiment.distances)):
        if not any(tmp in a for tmp in SKIP):
            new_names.append(a)
    print(len(new_names))

    distMatrix = np.zeros([len(new_names), len(new_names)])
    for i, a in enumerate(new_names):
        for j, b in enumerate(new_names):
            if a != b:
                distMatrix[i][j] = experiment.distances[a][b]

    # Zd = linkage(ssd.squareform(distMatrix), method="complete")
    # cld = fcluster(Zd, 500, criterion='distance').reshape(len(new_names), 1)

    Zd = linkage(ssd.squareform(distMatrix), method="complete")
    cld = fcluster(Zd, 12, criterion='maxclust').reshape(len(new_names), 1)

    clusters = {}
    for i, name in enumerate(new_names):
        clusters[name] = cld[i][0]
    for name in experiment.coordinates:
        if name not in clusters:
            clusters[name] = 0
    return {'value': clusters}


def clustering_kmeans(experiment, num_clusters=20):
    from sklearn.cluster import KMeans

    points = list(experiment.coordinates.values())

    kmeans = KMeans(n_clusters=num_clusters)

    kmeans.fit(points)

    y_km = kmeans.fit_predict(points)

    # plt.scatter(points[y_km == 0, 0], points[y_km == 0, 1], s=100, c='red')
    # plt.scatter(points[y_km == 1, 0], points[y_km == 1, 1], s=100, c='black')
    # plt.scatter(points[y_km == 2, 0], points[y_km == 2, 1], s=100, c='blue')
    # plt.scatter(points[y_km == 3, 0], points[y_km == 3, 1], s=100, c='cyan')

    # all_distances = []
    # for a,b in combinations(election.distances, 2):
    #     all_distances.append([a, b, election.distances[a][b]])
    # all_distances.sort(key=lambda x: x[2])
    #
    # clusters = {a: None for a in election.distances}
    # num_clusters = 0
    # for a,b,dist in all_distances:
    #     if clusters[a] is None and clusters[b] is None:
    #         clusters[a] = num_clusters
    #         clusters[b] = num_clusters
    #         num_clusters += 1
    #     elif clusters[a] is None and clusters[b] is not None:
    #         clusters[a] = clusters[b]
    #     elif clusters[a] is not None and clusters[b] is None:
    #         clusters[b] = clusters[a]
    clusters = {}
    for i, name in enumerate(experiment.coordinates):
        clusters[name] = y_km[i]
    return {'value': clusters}


def id_vs_un(experiment, election_ids, feature_params=None):
    results = {}
    experiment.distances['ID']['ID'] = 0
    experiment.distances['UN']['UN'] = 0
    for election_id in election_ids:
        distance_to_id = experiment.distances['ID'][election_id]
        distance_to_un = experiment.distances['UN'][election_id]
        if distance_to_id > distance_to_un:
            results[election_id] = 1
        elif distance_to_id < distance_to_un:
            results[election_id] = 0
        else:
            results[election_id] = 0.5
    return results


def an_vs_st(experiment, election_ids, feature_params=None):
    results = {}
    experiment.distances['AN']['AN'] = 0
    experiment.distances['ST']['ST'] = 0
    for election_id in election_ids:
        distance_to_an = experiment.distances['AN'][election_id]
        distance_to_st = experiment.distances['ST'][election_id]
        if distance_to_an > distance_to_st:
            results[election_id] = 1
        elif distance_to_an < distance_to_st:
            results[election_id] = 0
        else:
            results[election_id] = 0.5
    return results



#!/usr/bin/env python

from mapel.elections.cultures.group_separable import get_gs_caterpillar_vectors
from mapel.elections.cultures.single_peaked_matrices import get_walsh_vectors, \
    get_conitzer_vectors
from mapel.elections.cultures.single_crossing import get_single_crossing_vectors

from mapel.elections.objects.Election import Election
from mapel.elections.objects.OrdinalElectionExperiment import OrdinalElectionExperiment

from mapel.elections.cultures.mallows import get_mallows_matrix

import os
import csv


def prepare_matrices(experiment_id):
    """ compute positionwise matrices and
    is_exported them in the /matrices folder """
    experiment = OrdinalElectionExperiment(experiment_id)

    path = os.path.join(os.getcwd(), "election", experiment_id, "matrices")
    for file_name in os.listdir(path):
        os.remove(os.path.join(path, file_name))

    for election_id in experiment.elections:
        matrix = experiment.elections[election_id].votes_to_positionwise_matrix()
        file_name = election_id + ".csv"
        path = os.path.join(os.getcwd(), "election", experiment_id,
                            "matrices", file_name)

        with open(path, 'w', newline='') as csv_file:

            writer = csv.writer(csv_file, delimiter=';')
            header = [str(i) for i in range(experiment.elections[election_id].num_candidates)]
            writer.writerow(header)
            for row in matrix:
                writer.writerow(row)


def generate_positionwise_matrix(model_id=None, num_candidates=None,
                                 num_voters=100, params=None):

    if model_id == 'conitzer_matrix':
        vectors = get_conitzer_vectors(num_candidates)
    elif model_id == 'walsh_matrix':
        vectors = get_walsh_vectors(num_candidates)
    elif model_id == 'single-crossing_matrix':
        vectors = get_single_crossing_vectors(num_candidates)
    elif model_id == 'gs_caterpillar_matrix':
        vectors = get_gs_caterpillar_vectors(num_candidates)
    elif model_id == 'norm-mallows_matrix':
        return get_mallows_matrix(num_candidates, params)

    return vectors.transpose()


def get_positionwise_matrix(votes):
    election = Election("virtual", "virtual", votes=votes)
    return election.votes_to_positionwise_matrix()



import random
import numpy as np

from prefsampling.ordinal import single_peaked_conitzer as generate_ordinal_sp_conitzer_votes
from prefsampling.ordinal import single_peaked_walsh as generate_ordinal_sp_walsh_votes


def generate_ic_party(num_voters: int = None, params: dict = None) -> list:
    """ Return: party votes from Impartial Culture"""
    num_parties = params['num_parties']
    party_size = params['num_winners']

    votes = np.zeros([num_voters, num_parties], dtype=int)

    for j in range(num_voters):
        votes[j] = np.random.permutation(num_parties)

    new_votes = [[] for _ in range(num_voters)]
    for i in range(num_voters):
        for j in range(num_parties):
            for w in range(party_size):
                _id = votes[i][j] * party_size + w
                new_votes[i].append(_id)
    return new_votes



def generate_sp_party(model=None, num_voters=None, num_candidates=None, params=None) -> np.ndarray:
    candidates = [[] for _ in range(num_candidates)]
    _ids = [i for i in range(num_candidates)]

    for j in range(params['num_parties']):
        for w in range(params['num_winners']):
            _id = j * params['num_winners'] + w
            candidates[_id] = [np.random.normal(params['party'][j][0], params['var'])]

    mapping = [x for _, x in sorted(zip(candidates, _ids))]

    if model == 'conitzer_party':
        votes = generate_ordinal_sp_conitzer_votes(num_voters=num_voters,
                                                   num_candidates=num_candidates)
    elif model == 'walsh_party':
        votes = generate_ordinal_sp_walsh_votes(num_voters=num_voters,
                                                num_candidates=num_candidates)
    for i in range(num_voters):
        for j in range(num_candidates):
            votes[i][j] = mapping[votes[i][j]]

    return votes


def generate_approval_exp_partylist_votes(num_voters=None, num_candidates=None, params=None):
    if params is None:
        params = {}

    num_groups = params.get('g', 5)
    exp = params.get('experiment', 2.)

    sizes = np.array([1. / exp ** (i + 1) for i in range(num_groups)])
    sizes = sizes / np.sum(sizes)
    party_votes = np.random.choice([i for i in range(num_groups)], num_voters, p=sizes)

    party_size = int(num_candidates / num_groups)
    votes = []

    for i in range(num_voters):
        shift = party_votes[i] * party_size
        vote = set([int(c + shift) for c in range(party_size)])
        votes.append(vote)

    return votes



#!/usr/bin/env python
# UNDER CONSTRUCTION #

import os
from collections import Counter
import random as rand

import numpy as np


# GENERATE
def generate_preflib_election(experiment=None, model=None, name=None,
                              num_voters=None, num_candidates=None, folder=None,
                              selection_method='random'):
    """ main function: generate elections"""

    votes = generate_votes_preflib(model, selection_method=selection_method,
                                   num_voters=num_voters, num_candidates=num_candidates,
                                   folder=folder)

    path = os.path.join("experiments", experiment.experiment_id, "elections", name + ".soc")
    file_ = open(path, 'w')

    file_.write(str(num_candidates) + "\n")

    for i in range(num_candidates):
        file_.write(str(i) + ', c' + str(i) + "\n")

    c = Counter(map(tuple, votes))
    counted_votes = [[count, list(row)] for row, count in c.items()]
    counted_votes = sorted(counted_votes, reverse=True)

    file_.write(str(num_voters) + ', ' + str(num_voters) + ', ' + str(len(counted_votes)) + "\n")

    for i in range(len(counted_votes)):
        file_.write(str(counted_votes[i][0]) + ', ')
        for j in range(num_candidates):
            file_.write(str(counted_votes[i][1][j]))
            if j < num_candidates - 1:
                file_.write(", ")
            else:
                file_.write("\n")

    file_.close()


# REAL
def generate_votes_preflib(model, num_candidates=None, num_voters=None,  folder=None,
                           selection_method='borda', original=False):
    """ Generate votes based on elections from Preflib """

    long_name = str(model)
    file_name = '_real_data/' + folder + '/' + long_name + '.txt'
    file_votes = open(file_name, 'r')
    original_num_voters = int(file_votes.readline())
    if original_num_voters == 0:
        return [0, 0, 0]
    original_num_candidates = int(file_votes.readline())
    choice = [x for x in range(original_num_voters)]
    np.random.shuffle(choice)

    votes = np.zeros([num_voters, original_num_candidates], dtype=int)
    original_votes = np.zeros([original_num_voters, original_num_candidates], dtype=int)

    for j in range(original_num_voters):
        value = file_votes.readline().strip().split(',')
        for k in range(original_num_candidates):
            original_votes[j][k] = int(value[k])

    file_votes.close()
    # print(model, len(original_votes), len(np.unique(original_votes, axis=0)))
    if original:
        return np.array(original_votes)

    for j in range(num_voters):
        r = np.random.randint(0, original_num_voters - 1)
        for k in range(original_num_candidates):
            votes[j][k] = original_votes[r][k]

    for i in range(num_voters):
        if len(votes[i]) != len(set(votes[i])):
            print('wrong data')

    # REMOVE SURPLUS CANDIDATES
    if num_candidates < original_num_candidates:
        new_votes = []

        # NEW 17.12.2020
        if selection_method == 'random':
            selected_candidates = [j for j in range(original_num_candidates)]
            np.random.shuffle(selected_candidates)
            selected_candidates = selected_candidates[0:num_candidates]
        elif selection_method == 'borda':
            scores = get_borda_scores(original_votes, original_num_voters, original_num_candidates)
            order_by_score = [x for _, x in
                              sorted(zip(scores, [i for i in range(original_num_candidates)]),
                                     reverse=True)]
            selected_candidates = order_by_score[0:num_candidates]
        elif selection_method == 'freq':
            freq = import_freq(model)
            selected_candidates = freq[0:num_candidates]
        else:
            raise NameError('No such selection method!')

        mapping = {}
        for j in range(num_candidates):
            mapping[str(selected_candidates[j])] = j
        for j in range(num_voters):
            vote = []
            for k in range(original_num_candidates):
                cand = votes[j][k]
                if cand in selected_candidates:
                    vote.append(mapping[str(cand)])
            if len(vote) != len(set(vote)):
                print(vote)
            new_votes.append(vote)
        return np.array(new_votes)
    else:
        return np.array(votes)


def import_freq(elections_model):
    path = '_real_data/freq/' + elections_model + '.txt'
    with open(path, 'r', newline='') as txt_file:
        line = txt_file.readline().strip().split(',')
        line = line[0:len(line) - 1]
        for i in range(len(line)):
            line[i] = int(line[i])
    return line


def get_borda_scores(votes, num_voters, num_candidates):
    scores = [0 for _ in range(num_candidates)]
    for i in range(num_voters):
        for j in range(num_candidates):
            scores[votes[i][j]] += num_candidates - j - 1

    return scores


def prepare_preflib_family(experiment=None, model=None, family_id=None,
                           params=None) -> list:
    # NEEDS UPDATE #

    selection_method = 'random'
    keys = []

    # list of IDs larger than 10
    if model == 'irish':
        folder = 'irish_s1'
        # folder = 'irish_f'
        ids = [1, 3]
    elif model == 'glasgow':
        folder = 'glasgow_s1'
        # folder = 'glasgow_f'
        ids = [2, 3, 4, 5, 6, 7, 8, 9, 11, 13, 16, 19, 21]
    elif model == 'formula':
        folder = 'formula_s1'
        # 17 races or more
        ids = [17, 35, 37, 40, 41, 42, 44, 45, 46, 47, 48]
    elif model == 'skate':
        folder = 'skate_ic'
        # 9 judges
        ids = [1, 2, 3, 4, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20,
               25, 26, 27, 28, 29, 30, 31, 32, 33, 34,
               35, 36, 37, 38, 39, 40, 41, 42, 43, 44, 47, 48]
    elif model == 'sushi':
        folder = 'sushi_ff'
        ids = [1]
    elif model == 'grenoble':
        folder = 'grenoble_ff'
        ids = [1]
    elif model == 'tshirt':
        folder = 'tshirt_ff'
        ids = [1]
    elif model == 'cities_survey':
        folder = 'cities_survey_s1'
        ids = [1, 2]
    elif model == 'aspen':
        folder = 'aspen_s1'
        ids = [1]
    elif model == 'marble':
        folder = 'marble_ff'
        ids = [1, 2, 3, 4, 5]
    elif model == 'cycling_tdf':
        folder = 'cycling_tdf_s1'
        # ids = [e for e in range(1, 69+1)]
        selection_method = 'random'
        ids = [14, 15, 16, 17, 18, 19, 20, 21, 23, 24, 25, 26]
    elif model == 'cycling_gdi':
        folder = 'cycling_gdi_s1'
        ids = [i for i in range(2, 23 + 1)]
    elif model == 'ers':
        folder = 'ers_s1'
        # folder = 'ers_f'
        # 500 voters or more
        ids = [3, 9, 23, 31, 32, 33, 36, 38, 40, 68, 77, 79, 80]
    elif model == 'ice_races':
        folder = 'ice_races_s1'
        # 80 voters or more
        ids = [4, 5, 8, 9, 15, 20, 23, 24, 31, 34, 35, 37, 43, 44, 49]
    else:
        ids = []

    ctr = 0
    rand_ids = rand.choices(ids, k=experiment.families[family_id].size)
    for ri in rand_ids:
        name = family_id + '_' + str(ctr)
        tmp_election_type = model + '_' + str(ri)

        generate_preflib_election(
            experiment=experiment, model=tmp_election_type,
            name=name,
            num_voters=experiment.families[family_id].num_voters,
            num_candidates=experiment.families[family_id].num_candidates,
            folder=folder, selection_method=selection_method)
        ctr += 1

        keys.append(name)
    return keys


# MATRICES
def get_sushi_vectors():
    return np.array(get_sushi_matrix()).transpose()


def get_sushi_matrix():
    return [[0.11, 0.0808, 0.0456, 0.1494, 0.109, 0.0412, 0.3426, 0.0226, 0.0072, 0.0916],
            [0.1058, 0.1546, 0.0644, 0.1302, 0.1304, 0.0358, 0.2056, 0.051, 0.0132, 0.109],
            [0.1138, 0.1594, 0.0884, 0.096, 0.1266, 0.0548, 0.1276, 0.0874, 0.0246, 0.1214],
            [0.109, 0.1474, 0.1048, 0.0754, 0.1182, 0.068, 0.0862, 0.1168, 0.034, 0.1402],
            [0.1004, 0.1376, 0.129, 0.061, 0.0924, 0.0888, 0.0666, 0.1394, 0.047, 0.1378],
            [0.1042, 0.116, 0.1348, 0.0632, 0.0846, 0.0954, 0.048, 0.1528, 0.0804, 0.1206],
            [0.097, 0.0832, 0.1324, 0.0568, 0.079, 0.1306, 0.0422, 0.1682, 0.1094, 0.1012],
            [0.0982, 0.0636, 0.1214, 0.0638, 0.0728, 0.1546, 0.0376, 0.1396, 0.166, 0.0824],
            [0.0836, 0.0352, 0.1048, 0.1084, 0.101, 0.1692, 0.0236, 0.0958, 0.2222, 0.0562],
            [0.078, 0.0222, 0.0744, 0.1958, 0.086, 0.1616, 0.02, 0.0264, 0.296, 0.0396]]


def generate_preflib_votes(culture_id=None, num_candidates=None, num_voters=None, params=None):
    model = f'{culture_id}_{params["id"]}'
    votes = generate_votes_preflib(model, num_candidates, num_voters, params['folder'])
    return np.array(votes)



import csv
import random as rand


def generate_approval_grenoble_votes(num_voters=None, num_candidates=None,
                                         params=None):

    name = 'Grenoble_2017'
    path = f'source/{name}.csv'

    votes = []
    with open(path, 'r', newline='') as csv_file:
        reader = csv.DictReader(csv_file, delimiter=';')

        for row in reader:
            vote = set()
            vote_raw = list(row.values())[1:12]
            for i, value in enumerate(vote_raw):
                if value == '1':
                    vote.add(i)
            if vote:
                votes.append(vote)

    return rand.sample(votes, num_voters)


def generate_approval_field_votes(num_voters=None, num_candidates=None, params=None):

    folder = params['folder']
    name = params['name']

    if folder == 'french_2017':

        path = f'source/{folder}/{name}.csv'
        votes = []
        with open(path, 'r', newline='') as csv_file:
            reader = csv.DictReader(csv_file, delimiter=';')

            for row in reader:
                vote = set()
                vote_raw = list(row.values())[1:12]
                for i, value in enumerate(vote_raw):
                    if value == '1':
                        vote.add(i)
                if vote:
                    votes.append(vote)

    elif folder == 'french_2002':

        path = f'source/{folder}/{name}.toc'
        votes = []

        with open(path, 'r', newline='') as toc_file:

            num_candidates = int(toc_file.readline())
            for _ in range(num_candidates):
                toc_file.readline()

            line = toc_file.readline().rstrip("\n").split(',')
            total_num_voters = int(line[0])
            num_options = int(line[2])

            votes = []

            for j in range(num_options):
                line = toc_file.readline().strip()
                quantity = int(line.split(',')[0])
                line = line.split('{')
                if len(line) == 2:
                    raw_vote = set(line[0].split(',')[1])
                elif len(line) == 3:
                    raw_vote = set(line[1].replace('{', '').replace('}', '').split(',')[0:-1])

                vote = set()
                if raw_vote != {''}:
                    for c in raw_vote:
                        vote.add(int(c)-1)

                for _ in range(quantity):
                    votes.append(vote)

    return rand.sample(votes, num_voters)

